"""
Author: Amanda Beck <ambeck@mit.edu>
        Mingjian He <mh1@stanford.edu>
"""

import numpy as np
from numpy import random
import math
from scipy.stats import chi2
from spectrum import aryule, arma2psd, arburg
import matplotlib.pyplot as plt
import warnings
from kneed import KneeLocator
from ..multitaper import fast_psd_multitaper
from ..multitaper.multitaper_spectrogram_python import multitaper_spectrogram as mtm
from ..basic_models import OscillatorModel as Osc
from ..basic_models import AutoRegModel as Arn


def innovations_wrapper(iter_osc, iter_num, plot_all_poles=True, ax_limit=None, figsize=(8, 6), horizontal=False):
    """ Plot the innovation spectrum of the fitted model at a specific iteration """
    y = iter_osc.added_osc[0].y
    osc = iter_osc.fitted_osc[iter_num]
    innovations, _ = get_innovations(osc, y=y)

    add_freq = iter_osc.added_osc[iter_num + 1].freq[0] if iter_num + 1 < len(iter_osc.added_osc) else None
    add_radius = iter_osc.added_osc[iter_num + 1].a[0] if iter_num + 1 < len(iter_osc.added_osc) else None

    a, p = fit_ar(np.squeeze(innovations), ar_order=iter_osc.osc_range * 2 - 1)
    fig, axlim = plot_innovations(osc, y, innovations, a, p, add_freq, add_radius, plot_all_poles=plot_all_poles,
                                  ax_limit=ax_limit, figsize=figsize, horizontal=horizontal, counter=iter_num)

    return fig, axlim


def get_innovations(osc, y=None):
    """ Calculate innovations / one-step prediction error """
    y = osc.y if y is None else y
    kalman_out = osc.dejong_filt_smooth(y=y, return_dict=True)
    x_pred = kalman_out['x_t_tmin1'][:, 1:]
    y_pred = osc.G @ x_pred
    return (y - y_pred).squeeze(), kalman_out['logL'].sum()


def fit_ar(y, ar_order, burg_flag=False):
    """ Fit AR model to a time series """
    if burg_flag:  # use the Burg method
        a, p, _ = arburg(y, ar_order)
    else:  # default is Yule Walker algorithm
        a, p, _ = aryule(y, ar_order)
    return a, p  # first output is the negative of AR coefficients in y_t = \sum a*y_{t-1} + e_t


def get_ar_psd(fs, a, p, ar_hz=None, df=0.01):
    """ Get the PSD spectrum of a fitted AR model """
    if ar_hz is None:
        nfft = int(2 ** np.ceil(np.log2(fs / df)))  # nfft to achieve a [< df] sampling resolution of frequencies
        ar_hz = np.arange(0, fs / 2, fs / nfft)  # one-sided frequency 0 to Nyquist
        assert ar_hz.size == nfft // 2, 'Incorrect nfft in the AR theoretical spectrum.'
    ar_psd = arma2psd(A=a, rho=p, NFFT=2 * ar_hz.size, sides='centerdc')[ar_hz.size:] * 2  # make spectrum one-sided
    return ar_psd, ar_hz


def initialize_allosc(fs, y, R, osc_range=7, drop_y=False, fill_up=True):
    """ Initialize all new oscillators using spectral decomposition """
    a, p = fit_ar(np.squeeze(y), ar_order=osc_range * 2 - 1)
    r = np.roots(np.concatenate([[1], a]))

    # AR estimated root frequencies and radii
    root_freqs = np.abs(np.arctan2(np.imag(r), np.real(r)) * fs / 2 / np.pi)  # force to positive frequencies
    root_freqs = np.mod(root_freqs, fs / 2)  # map Nyquist to dc frequency
    root_radii = np.abs(r)  # compute the magnitude of complex roots

    # spectral decomposition to estimate root specific variances
    eig, eigvec = np.linalg.eig(Arn(coeff=-a, sigma2=p).F)
    assert np.all(r == eig), 'eigenvalues do not equal to complex roots.'
    c = eigvec[0, :] * np.linalg.inv(eigvec)[:, 0]  # sum of c is 1
    root_sigma2 = np.real(p * c * c.conj())  # the variance in each mode when observed alone

    # sort by descending root variance
    unique_freqs = np.unique(root_freqs)
    unique_sigma2 = np.array([np.sum(root_sigma2[root_freqs == x]) for x in unique_freqs])
    add_indices = np.argsort(unique_sigma2)[::-1]

    # lower bound sigma2 to be above 0 after sorting
    unique_sigma2[unique_sigma2 <= 0] = np.finfo(float).eps

    # create a list of oscillators to be added sequentially
    added_osc = []
    for ii in range(len(add_indices)):
        add_freq = unique_freqs[add_indices[ii]]
        add_radius = max(root_radii[root_freqs == add_freq])
        add_sigma2 = unique_sigma2[add_indices[ii]]
        if ii == 0 and not drop_y:
            added_osc.append(Osc(y=y, a=add_radius, freq=max([add_freq, 0.1]), sigma2=add_sigma2, Fs=fs, R=R))
        else:
            added_osc.append(Osc(a=add_radius, freq=max([add_freq, 0.1]), sigma2=add_sigma2, Fs=fs))

    # fill up to osc_range oscillators to be added
    if fill_up:
        while len(added_osc) < osc_range:
            o1 = added_osc[0].copy()
            _ = [o1.append(x) for x in added_osc[1:]]
            residual = o1.y - o1.G @ o1.dejong_filt_smooth(return_dict=True)['x_t_tmin1'][:, 1:]
            more_osc = initialize_allosc(fs, residual, R, osc_range=osc_range, drop_y=True, fill_up=False)
            for ii in range(min([osc_range - len(added_osc), len(more_osc)])):
                added_osc.append(more_osc[ii])
        assert len(added_osc) == osc_range, 'Incorrect number of oscillators during decomposition.'

    return added_osc


def initialize_newosc(fs, innovations, existing_freqs=None, freq_res=1, ar_order=13, burg_flag=False, sigma2='eig'):
    """ Initialize the new oscillator to be added """
    a, p = fit_ar(np.squeeze(innovations), ar_order=ar_order, burg_flag=burg_flag)
    r = np.roots(np.concatenate([[1], a]))
    root_freqs = np.abs(np.arctan2(np.imag(r), np.real(r)) * fs / 2 / np.pi)  # force to positive frequencies
    root_radii = np.abs(r)  # compute the magnitude of complex roots

    # within [freq_res] Hz of existing frequencies is considered as duplicated
    root_freqs_sel = [True if existing_freqs is None else min(abs(x - existing_freqs)) >= freq_res for x in root_freqs]
    root_freqs_sel = np.invert(root_freqs_sel) if not any(root_freqs_sel) else root_freqs_sel  # all root_freqs overlap

    # look for the root with the largest AR theoretical PSD scaled by radius
    ar_psd, ar_hz = get_ar_psd(fs, a, p)
    root_psd = np.array([ar_psd[np.argmin(abs(freq - ar_hz))] * radius for freq, radius in zip(root_freqs, root_radii)])
    sel_idx = np.argmax(root_psd[root_freqs_sel])
    add_freq = max([root_freqs[root_freqs_sel][sel_idx], 0.1])  # lower bound added frequency to 0.1 Hz
    add_radius = root_radii[root_freqs_sel][sel_idx]

    # compute the corresponding sigma2 for the new oscillator
    # via eigendecomposition and transform of the AR process white noise variance
    eig, eigvec = np.linalg.eig(Arn(coeff=-a, sigma2=p).F)  # eig is equivalent to complex roots in r
    assert np.all(r == eig), 'eigenvalues do not equal to complex roots.'
    c = eigvec[0, :] * np.linalg.inv(eigvec)[:, 0]  # sum of c is 1
    root_sigma2 = np.real(p * c * c.conj())  # the variance in each mode when observed alone
    add_sigma2_eig = np.sum(root_sigma2[root_freqs == root_freqs[root_freqs_sel][sel_idx]])  # sum across pairs of roots

    # via empirical matching to the fitted AR PSD
    add_psd = root_psd[root_freqs_sel][sel_idx]  # this psd is already scaled by radius
    w = Osc.hz_to_rad(add_freq, fs)
    z = np.exp(1j * Osc.hz_to_rad(root_freqs[root_freqs_sel][sel_idx], fs))
    A = (1 - 2 * add_radius ** 2 * np.cos(w) ** 2 + add_radius ** 4 * np.cos(2 * w)) / (
            add_radius * (add_radius ** 2 - 1) * np.cos(w))
    b = 0.5 * (A - 2 * add_radius * np.cos(w) + np.sqrt((A - 2 * add_radius * np.cos(w)) ** 2 - 4))
    V = 0.5 * add_psd * np.abs(1 - 2 * add_radius * np.cos(w) * z + add_radius ** 2 * z ** 2
                               ) ** 2 / np.abs(1 + b * z) ** 2
    add_sigma2_psd = -V * b / (add_radius * np.cos(w))

    if sigma2 == 'avg':
        add_sigma2 = np.mean([add_sigma2_eig, add_sigma2_psd])
    elif sigma2 == 'eig':
        add_sigma2 = add_sigma2_eig
    else:
        add_sigma2 = add_sigma2_psd

    return add_freq, add_radius, add_sigma2, root_freqs, root_radii, a, p


def aic_calc(osc, ll):
    """ Calculate AIC """
    if osc.dc_idx is None:
        k = osc.ncomp * 3 + 1  # each oscillator has 3 parameters + obs noise variance
    else:
        k = osc.ncomp * 3  # each osc has 3 parameters, each dc has 2, + obs noise variance
    return 2 * k - 2 * ll


def aic_weights(aic_list):
    """
    Calculate AIC weights.
    Weights represent each model's likelihood relative to the set of models.
    They can be used as a measure of model probability
    """
    deltas = (aic_list - np.min(aic_list)) / 2
    return np.exp(-deltas) / np.sum(np.exp(-deltas))


def ebic_calc(osc, ll, osc_range, gamma=0.5):
    """
    Calculate extended Bayesian Information Criterion (EBIC).
    When gamma=0, EBIC is equivalent to BIC

    Reference:
        Chen, J., & Chen, Z. (2008). Extended Bayesian information criteria
        for model selection with large model spaces. Biometrika, 95(3), 759-771.
    """
    if osc.dc_idx is None:
        k = osc.ncomp * 3 + 1  # each oscillator has 3 parameters + obs noise variance
    else:
        k = osc.ncomp * 3  # each osc has 3 parameters, each dc has 2, + obs noise variance

    return k * np.log(osc.ntime) - 2 * ll + 2 * gamma * np.log(math.comb(osc_range, osc.ncomp))


def get_knee(ll_vec):
    """ Use 'kneed' python package to find a knee on the log likelihoods """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message='No knee/elbow found', category=UserWarning)
        kneedle = KneeLocator(range(len(ll_vec)), ll_vec, S=1.0,
                              curve="concave", direction="increasing", online=False)  # stop at the first knee
    knee_index = kneedle.knee if kneedle.knee is not None else len(ll_vec) - 1  # if no knee found, output last index
    return knee_index  # Returns int, index of model at knee


def get_last_increase(sequence):
    """ Find the index before the first decrease in a sequence """
    sequence_diff = np.diff(sequence)
    for idx in range(len(sequence_diff)):
        if sequence_diff[idx] <= 0:
            return idx
    return len(sequence) - 1  # return index of the last value in sequence


def simulate_matsuda(param_list, R=1, Fs=100, T=10):
    """
    Simulate time series from matsuda oscillators (Matsuda and Komaki 2017)
        param_list = array of dictionaries, one for each oscillation with the following entries:
            a = scalar damping factor
            q = scalar state noise covariance
            f = scalar frequency
        R = scalar observation noise
        Fs = scalar sampling frequency
        T = scalar time length in seconds
    """
    # total number of samples in the time series
    N = T * Fs

    # verify all required keys are available
    for ii in range(len(param_list)):
        for key, name in zip(['a', 'q', 'f'], ['Damping factor', 'State noise cov', 'Frequency']):
            assert key in param_list[ii], '%s is missing from oscillation %d' % (name, ii)

    # generate the time series with an extra t=0 point that will get removed during return
    y = np.zeros((1, N + 1))
    for params in param_list:
        w = params['f'] * 2 * np.pi / Fs
        noise = random.normal(0, params['q'] ** (1 / 2), size=(2, N))
        F = Osc.get_rot_mat(params['a'], w)
        params['x_osc'] = np.zeros((2, N + 1))
        for ii in range(N):
            params['x_osc'][:, ii + 1] = noise[:, ii] + F @ params['x_osc'][:, ii]
        y += params['x_osc'][0, :]
        params['R'] = R
        params['Fs'] = Fs

    # add observation noise
    noise_meas = random.normal(0, R ** (1 / 2), size=(1, N + 1))
    y += noise_meas

    return y.squeeze()[1:], param_list, noise_meas


def sim_to_osc_object(y, param_list):
    """ Reorganize a list of simulated oscillators into an OscillatorModel object """
    assert np.all([param_list[0]['R'] == ii['R'] for ii in param_list]), \
        'Not all observation noises are the same. Check R'
    assert np.all([param_list[0]['Fs'] == ii['Fs'] for ii in param_list]), \
        'Not all sampling frequencies are the same. Check Fs'

    osc_init = {'y': y, 'a': [ii['a'] for ii in param_list], 'freq': [ii['f'] for ii in param_list],
                'sigma2': [ii['q'] for ii in param_list],
                'Fs': param_list[0]['Fs'], 'R': param_list[0]['R'], 'add_dc': False}
    osc_out = Osc(**osc_init)
    x_osc = np.vstack([param_list[ii]['x_osc'] for ii in range(len(param_list))])
    return osc_out, x_osc[:, 1:]


def plot_param(iterosc, title, sim_osc=None):
    """ Plotting parameters: AIC, R, frequency """
    osc_num = len(iterosc.AIC)
    osc_choice = np.argmin(iterosc.AIC)

    f, ax = plt.subplots(5, 1, sharex='all', figsize=(4, 14))
    ax[0].set_title(title)
    _ = [ii.grid(axis='y', which='both') for ii in ax]
    _ = [ii.axvspan(osc_choice - 0.25, osc_choice + 0.25, alpha=0.3, color='yellow') for ii in ax]

    for (ii, label) in zip(range(5), ['AIC weights', 'Obs noise (R)',
                                      'Frequency', 'Damping Factor', 'State noise (Q)']):
        ax[ii].set_ylabel(label)

    AIC_weights = aic_weights(iterosc.AIC)
    ax[0].scatter(range(osc_num), AIC_weights)
    ax[1].scatter(range(osc_num), [ii.R[0] for ii in iterosc.fitted_osc])
    for ii in range(osc_num):
        value_on(ax[0], ii, AIC_weights[ii])
        value_on(ax[1], ii, iterosc.fitted_osc[ii].R[0])

    colors = ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple']
    for ii in range(len(iterosc.fitted_osc)):
        osc = iterosc.fitted_osc[ii]
        for jj in range(ii + 1):
            ax[2].scatter(ii, np.abs(osc.w[jj]) * osc.Fs / (2 * np.pi), color=colors[jj])
            value_on(ax[2], ii, np.abs(osc.w[jj]) * osc.Fs / (2 * np.pi))
            ax[3].scatter(ii, osc.a[jj], color=colors[jj])
            value_on(ax[3], ii, osc.a[jj])
            ax[4].scatter(ii, osc.sigma2[jj], color=colors[jj])
            value_on(ax[4], ii, osc.sigma2[jj])

    ax[-1].set_xticks(np.arange(osc_num), [str(ii) for ii in np.arange(osc_num) + 1])
    ax[-1].set_xlabel('Number of Oscillations')
    ax[-1].margins(x=0.25)

    if sim_osc is not None:
        ax[1].axhline(sim_osc.R[0], linestyle=':')
        for ii in range(sim_osc.ncomp):
            ax[2].axhline(sim_osc.w[ii] * sim_osc.Fs / (2 * np.pi), linestyle=':', color=colors[ii])
            ax[3].axhline(sim_osc.a[ii], linestyle=':', color=colors[ii])
            ax[4].axhline(sim_osc.sigma2[ii], linestyle=':', color=colors[ii])

    for ii in ax:
        ii.margins(y=0.25)
        ii.autoscale_view(tight=False)


def value_on(ax, current_osc, value):
    """ Add text of value at (x,y)=(current_osc,value) """
    ax.text(current_osc + 0.1, value, '%0.2f' % value)


def visualize_em(em_params, y, bw=None):
    """
    Plot tracked parameters from EM iterations.

    iterosc.all_params is a list of lists. The inner lists are one list per oscillator,
    with the Osc objects over the EM iterations. The inner list is em_params
    """
    fig, ax = plt.subplots(1, 1)
    Fs = em_params[0].Fs
    psd_mt, f_hz = fast_psd_multitaper(y, Fs, 0, Fs / 2, bw)
    ax.plot(f_hz, 10 * np.log10(psd_mt), color='black', label='observed data')
    cm = plt.get_cmap('coolwarm')
    for ii in range(len(em_params)):
        # noinspection PyProtectedMember
        h_i, f_th = em_params[ii]._theoretical_spectrum()
        for jj in range(len(h_i)):
            cl = int(ii / len(em_params) * 256)  # y further 0~Convert to 255
            ax.plot(f_th, 10 * np.log10(h_i[jj]), c=cm(cl))
    ax.legend()
    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('PSD (dB)')
    return fig


def plot_mtm(osc, frequency_range=None, time_bandwidth=2, window_params=(1, 0.05), df=0.1,
             detrend_opt='constant', ax=(None, None)):
    """ Plot multitaper spectrogram and mean spectrum """
    min_nfft = int(2 ** np.ceil(np.log2(osc.added_osc[0].Fs / df)))
    if ax[0] is not False:
        mt_spectrogram, stimes, sfreqs, (fig_spectrogram, ax0) = mtm(osc.added_osc[0].y, osc.added_osc[0].Fs,
                                                                     frequency_range=frequency_range,
                                                                     time_bandwidth=time_bandwidth,
                                                                     window_params=list(window_params),
                                                                     min_nfft=min_nfft, detrend_opt=detrend_opt,
                                                                     return_fig=True, ax=ax[0])
        ax0.set_xlabel('Time (Second)')
        ax0.set_title('Multitaper Spectrogram')
    else:
        mt_spectrogram, stimes, sfreqs = mtm(osc.added_osc[0].y, osc.added_osc[0].Fs,
                                             frequency_range=frequency_range,
                                             time_bandwidth=time_bandwidth,
                                             window_params=list(window_params),
                                             min_nfft=min_nfft, detrend_opt=detrend_opt,
                                             return_fig=False)
        fig_spectrogram = None

    if ax[1] is not False:
        if ax[1] is None:
            fig_spectrum, ax1 = plt.subplots(1, 1)
        else:
            ax1 = ax[1]
            fig_spectrum = ax1.get_figure()
        ax1.plot(sfreqs, 10 * np.log10(np.mean(mt_spectrogram, axis=1)))
        ax1.set_xlabel('Frequency (Hz)')
        ax1.set_ylabel('PSD (dB)')
        ax1.set_title('Mean Spectrum')
    else:
        fig_spectrum = None

    return fig_spectrogram, fig_spectrum


def plot_innovations(osc, y, innovations, a, p, add_freq=None, add_radius=None, reiterate=False, ax=None,
                     plot_all_poles=True, ax_limit=None, figsize=(8, 6), horizontal=False, bw=1, counter=0):
    """ Plot fitted OscillatorModel and the innovations / one-step prediction error for that model """
    psd_mt, fy_hz = fast_psd_multitaper(innovations, osc.Fs, 0, osc.Fs / 2, bw)
    psd_ar, _ = get_ar_psd(osc.Fs, a, p, ar_hz=fy_hz)
    psd_y, _ = fast_psd_multitaper(y.squeeze(), osc.Fs, 0, osc.Fs / 2, bw)

    if ax is None:
        if horizontal:
            fig, [ax0, ax1] = plt.subplots(1, 2, figsize=figsize)
        else:
            fig, [ax0, ax1] = plt.subplots(2, 1, sharex='all', figsize=figsize)
    else:
        assert len(ax) == 2, 'Need two axes for the innovations plot.'
        fig = ax[0].get_figure()
        ax0, ax1 = ax

    ax0.plot(fy_hz, 10 * np.log10(psd_y), color='black', label='observed data')

    h_th, f_th = osc.oscillator_spectra('theoretical')
    for ii in range(len(h_th)):
        ax0.plot(f_th, 10 * np.log10(h_th[ii]), label='osc %d fit, %0.2f Hz' % (ii + 1, osc.freq[ii]))
    ax0.legend(loc='upper right')
    ax0.set_ylabel('PSD (dB)')
    if reiterate:
        ax0.set_title('Re-iteration %d' % counter)
    else:
        ax0.set_title('Iteration %d' % counter)

    ax1.plot(fy_hz, 10 * np.log10(psd_mt), label='OSPE', color='tab:blue')
    ax1.plot(fy_hz, 10 * np.log10(psd_ar), label='AR fit', linestyle='dashed', color='tab:blue')
    if ax_limit is not None:
        ax1.set_ylim(ax_limit)

    ax1.set_xlabel('Frequency (Hz)')
    ax1.set_ylabel('PSD (dB)', color='tab:blue')
    ax1.tick_params(axis='y', labelcolor='tab:blue')

    ax1.set_title('One-Step Prediction Error (OSPE) - Iteration %d' % counter)
    if add_freq is not None:
        ax1.axvline(add_freq, color='black', linestyle='dashed',
                    label='frequency of new oscillator: %0.2f Hz' % add_freq)
    # ax1.axhline(20 * np.log10(b), color='black', label='observation noise estimate')

    ax2 = ax1.twinx()
    if plot_all_poles:
        r = np.roots(np.concatenate([[1], a]))
        root_freqs = np.abs(np.arctan2(np.imag(r), np.real(r)) * osc.Fs / 2 / np.pi)  # force to positive frequencies
        ax2.scatter(root_freqs, np.abs(r), color='tab:orange')
    if add_freq is not None and add_radius is not None:
        ax2.scatter(add_freq, add_radius, facecolor='tab:green', linewidth=2, label='pole for initialization')

    ax2.set_ylim([0, 1])
    ax2.set_xlabel('Frequency (Hz)')
    ax2.set_ylabel('Root (magnitude)', color='tab:green')
    ax2.tick_params(axis='y', labelcolor='tab:green')
    ax2.spines['left'].set_color('tab:blue')
    ax2.spines['right'].set_color('tab:green')
    ax2.spines['left'].set(lw=3)
    ax2.spines['right'].set(lw=3)

    lines, labels = ax1.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax1.legend(lines + lines2, labels + labels2, loc='lower center')

    fig.set_tight_layout(True)  # otherwise, the right y-label is slightly clipped

    return fig, ax1.get_ylim()


def plot_fit_line(fig, slope, intercept, ax=None):
    """ Add a linear fit line to an existing innovation plot figure """
    ax = ax or fig.axes[0]
    x_limits = ax.get_xlim()
    l1 = ax.plot(x_limits, [x * slope + intercept for x in x_limits], label='linear fit', color='black')
    ax.set_xlim(x_limits)

    lines = [x for x in ax.get_legend().legend_handles] + l1
    labels = [x.get_label() for x in lines]
    ax.legend(lines, labels, loc='lower center')

    return fig


def plot_acf(residual, Fs, ax=None):
    """ Plot the autocorrelation function of the residuals """
    from statsmodels.graphics.tsaplots import plot_acf

    if ax is None:
        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    else:
        assert len(ax) == 2, 'Need two axes for the acf plot.'
        fig = ax[0].get_figure()

    plot_acf(residual, lags=Fs, ax=ax[0])
    ax[0].set_xlabel('Lags (Samples)')
    ax[0].set_title(ax[0].get_title() + ' (Fs = {0:.1f} Hz)'.format(Fs))

    plot_acf(residual, lags=Fs, ax=ax[1])
    ax[1].set_xscale('log')
    ax[1].set_xlabel('Frequency (Hz)')
    ax[1].set_title(ax[1].get_title() + ' (Fs = {0:.1f} Hz)'.format(Fs))
    freqs = np.array([Fs, Fs/2, 30, 20, 12, 10, 8, 4, 2, 1])
    ax[1].set_xticks(Fs/freqs, ['{0:.0f}'.format(x) for x in freqs])
    ax[1].set_xticks([], minor=True)

    return fig


def plot_pacf(residual, Fs, ax=None):
    """ Plot the partial autocorrelation function of the residuals """
    from statsmodels.graphics.tsaplots import plot_pacf

    if ax is None:
        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    else:
        assert len(ax) == 2, 'Need two axes for the pacf plot.'
        fig = ax[0].get_figure()

    # if the residual is too short, the pacf will be unreliable
    lags = max(Fs/2, 10) if residual.size <= Fs * 2 else Fs  # at least 10 lags

    plot_pacf(residual, lags=lags, ax=ax[0])
    ax[0].set_xlabel('Lags (Samples)')
    ax[0].set_title(ax[0].get_title() + ' (Fs = {0:.1f} Hz)'.format(Fs))

    plot_pacf(residual, lags=lags, ax=ax[1])
    ax[1].set_xscale('log')
    ax[1].set_xlabel('Frequency (Hz)')
    ax[1].set_title(ax[1].get_title() + ' (Fs = {0:.1f} Hz)'.format(Fs))
    freqs = np.array([Fs, Fs/2, 30, 20, 12, 10, 8, 4, 2, 1])
    ax[1].set_xticks(Fs/freqs, ['{0:.0f}'.format(x) for x in freqs])
    ax[1].set_xticks([], minor=True)

    return fig


def plot_residual(osc, y, residual, fit_version='predicted', ax=None, bw=1):
    """
    Plot the best fitted OscillatorModel and the residuals from that model

    This function is similar to plot_innovations(), but it shows empirical PSD
    of fitted oscilaltors, and it shows the confidence interval of the residuals
    """
    psd_y, _ = fast_psd_multitaper(y.squeeze(), osc.Fs, 0, osc.Fs / 2, bw)
    psd_mt, fy_hz = fast_psd_multitaper(residual, osc.Fs, 0, osc.Fs / 2, bw)

    # Recompute the number of tapers used to form Chi-sqaured confidence intervals
    # this calculation is taken directly from the fast_psd_multitaper() function
    window_length = min(residual.size / osc.Fs, 10)  # max to 10s windows
    time_bandwidth = np.max([1., bw / 2 * window_length])
    num_tapers = math.floor(2 * time_bandwidth) - 1
    alpha = 0.95
    lower_bound = 10 * np.log10(2 * num_tapers / chi2.ppf(0.5 + alpha / 2, 2 * num_tapers))
    upper_bound = 10 * np.log10(2 * num_tapers / chi2.ppf(0.5 - alpha / 2, 2 * num_tapers))

    if ax is None:
        fig, [ax0, ax1] = plt.subplots(2, 1, sharex='all')
    else:
        assert len(ax) == 2, 'Need two axes for the model/residual plot.'
        fig = ax[0].get_figure()
        ax0, ax1 = ax

    ax0.plot(fy_hz, 10 * np.log10(psd_y), color='black', label='observed data')

    if fit_version == 'theoretical':
        h_th, f_th = osc.oscillator_spectra('theoretical')
    else:
        kalman_out = osc.kalman_filt_smooth(y, return_dict=True)
        if fit_version == 'predicted':
            x = kalman_out['x_t_tmin1'][:, 1:]
        elif fit_version == 'filtered':
            x = kalman_out['x_t_t'][:, 1:]
        elif fit_version == 'smoothed':
            x = kalman_out['x_t_n'][:, 1:]
        else:
            raise ValueError('Invalid fit_version. Choose from "predicted", "filtered", or "smoothed"')
        h_th, f_th = osc.oscillator_spectra('actual', x, bw)

    for ii in range(len(h_th)):
        ax0.plot(f_th, 10 * np.log10(h_th[ii]), label='osc %d fit, %0.2f Hz' % (ii + 1, osc.freq[ii]))
    ax0.legend(loc='upper right')
    ax0.set_ylabel('PSD (dB)')
    ax0.set_title(f'Selected Model Fit ({fit_version})')

    noise_level = np.squeeze((np.sum(osc.sigma2) + osc.R)) * 2  # adjust for the one-sided spectrum
    ax1.plot(fy_hz, 10 * np.log10(psd_mt / noise_level), label='residual / noises', color='tab:blue')

    ax1.axhline(0, color='gray', linestyle='dashed')
    ax1.axhline(lower_bound, color='gray')
    ax1.axhline(upper_bound, color='gray')
    xlimits = ax1.get_xlim()
    ax1.fill_between([ax1.get_xlim()[0], ax1.get_xlim()[1]], lower_bound, upper_bound,
                     facecolor='gray', alpha=0.25, label='95% CI')
    ax1.set_xlim(xlimits)

    ax1.set_xlabel('Frequency (Hz)')
    ax1.set_ylabel('PSD (dB)')
    ax1.set_title('Residual over estimated white noises')
    ax1.legend(loc='lower center')

    return fig



"""
Author: Mingjian He <mh1@stanford.edu>

decimate module contains methods for decimated state-space model learning
"""

from somata.basic_models import StateSpaceModel as Ssm
from somata.exact_inference import run_em
import numpy as np


class DecimatedModel(object):
    """
    DecimatedModel is an object class created to organize
    processing for decimated state-space model learning

    The structure of this class is very similar to the
    somata.switching.vb.VBSwitchModel class since both
    algorithms assume a set of parallel and independent
    state-space models in the hidden states.
    """
    ssm_array = None
    K = None
    T = None
    q = None
    logL = None

    def __init__(self, ssm_array):
        """
        Constructor method for DecimatedModel class
        :param ssm_array: an array of Ssm class objects with decimated observations
        """
        # The array length is the decimation factor
        self.K = len(ssm_array)

        # Verify that each Ssm object contains observed data
        for ii in range(self.K):
            assert ssm_array[ii].y is not None, 'No observed data contained in the ssm_array[%d] object.' % ii
            assert ssm_array[ii].ntime > 0, 'Observed data have length 0 in the ssm_array[%d] object.' % ii
            assert ssm_array[ii].nchannel > 0, 'Observed data have 0 channel in the ssm_array[%d] object.' % ii
        self.ssm_array = ssm_array

        # Get the total number of time points
        self.T = np.sum([ssm.ntime for ssm in ssm_array])

        # Get the number of channels that should be the same across objects
        qs = [ssm.nchannel for ssm in ssm_array]
        assert len(np.unique(qs)) == 1, 'Objects in ssm_array have observed data with different numbers of channels.'
        self.q = qs[0]

    def __repr__(self):
        """ Unambiguous and concise representation when calling DecimatedModel() """
        return 'Dec(' + str(self.K) + ')<' + hex(id(self))[-4:] + '>'

    def __str__(self):
        """ Helpful information when calling print(DecimatedModel()) """
        print_str = "Dec model " + str(self.ssm_array)
        return print_str

    def learn(self, keep_param=(), priors=None, max_iter=100, logL_thresh=1e-6, show_pbar=False):
        """
        Decimated learning of state-space models

        Perform EM learning by pooling sums of squares across
        parallel state-space models to update a single set
        of parameters for the lagged sequences that decimate
        the original time series with a high sampling rate.

        Inputs:
        :param keep_param: a tuple of strings for parameters to keep and not update in M step
        :param priors: a single dictionary specifying priors for all models, if None -> MLE
        :param max_iter: maximal number of EM iterations
        :param logL_thresh: threshold below which EM iteration stops
        :param show_pbar: show progress bar during EM iterations
        """
        # EM iterations
        self.logL = [float('-inf')] if self.logL is None else self.logL
        m_kwargs = {'priors': priors, 'keep_param': keep_param}
        em_log = run_em(self, m_kwargs=m_kwargs, max_iter=max_iter, stop_thresh=logL_thresh,
                        return_dict=True, show_pbar=show_pbar)
        return em_log

    def e_step(self, y=None):
        """
        E step in decimated oscillator learning, exposed for run_em()

        Using parallel Kalman (DeJong version) filtering and smoothing.

        Inputs:
        :param y: observed data (should be None only)
        """
        assert y is None, 'A single sequence of observed data should not be inputted in decimated oscillator learning.'

        x_t_n_all, P_t_n_all, P_t_tmin1_n_all, logL_all, *_ = Ssm.par_kalman(self.ssm_array, method='dejong',
                                                                             skip_check_observed=True)
        e_results = {'x_t_n_all': x_t_n_all, 'P_t_n_all': P_t_n_all, 'P_t_tmin1_n_all': P_t_tmin1_n_all}

        logL = float(np.sum([logL.sum() for logL in logL_all]))  # total logL summed across sequences
        stop_var = logL - self.logL[-1]  # logL should monotonically increase
        self.logL.append(logL)

        return e_results, stop_var

    def m_step(self, y=None, x_t_n_all=None, P_t_n_all=None, P_t_tmin1_n_all=None, priors=None, keep_param=()):
        """
        M step in decimated oscillator learning, exposed for run_em()

        Inputs:
        :param y: observed data (should be None only)
        :param x_t_n_all: a list of smoothed estimates (posterior) of state mean
        :param P_t_n_all: a list of smoothed estimates (posterior) of state conditional covariance
        :param P_t_tmin1_n_all: a list of smoothed estimates (posterior) of state lag1 conditional cross-covariance
        :param priors: a single dictionary specifying priors for all models, if None -> MLE
        :param keep_param: a tuple of strings for parameters to keep and not update
        """
        assert y is None, 'A single sequence of observed data should not be inputted in decimated oscillator learning.'

        # Compute sums of squares from each decimated sequence
        R_ss = np.zeros((self.q, self.q, self.K), dtype=np.float64)
        A, B, C, x_0_n, P_0_n = ([], [], [], [], [])
        for k in range(self.K):
            m_results = self.ssm_array[k].m_estimate(x_t_n=x_t_n_all[k], P_t_n=P_t_n_all[k],
                                                     P_t_tmin1_n=P_t_tmin1_n_all[k],
                                                     priors=priors, force_ABC=True, update_param=(), return_dict=True)
            # Store sum variables for joint estimations
            R_ss[:, :, k] = m_results['R_ss']
            A.append(m_results['A'])
            B.append(m_results['B'])
            C.append(m_results['C'])
            x_0_n.append(x_t_n_all[k][:, 0][:, None])
            P_0_n.append(P_t_n_all[k][:, :, 0])

        # TODO: an important assumption now is that there is only a single component in these ssm_array objects.
        #       We need to think about what to do when different components require different decimation rates.
        assert np.all([ssm.ncomp == 1 for ssm in self.ssm_array]), 'Only single component Ssm objects are supported.'

        # Obtain boolean flags of scopes of updates
        # noinspection PyProtectedMember
        update_FQ, update_mu0S0G = Ssm._m_estimate_scope(keep_param=keep_param)

        # Joint estimation of parameters
        if 'R' not in keep_param:
            R_ss = R_ss.sum(axis=2)
            # noinspection PyProtectedMember
            _ = [ssm._m_update_r(R_ss=R_ss, T=self.T, priors=priors) for ssm in self.ssm_array]

        # Call model specific _m_update_<param> methods
        if update_FQ:
            # Sum across decimated sequences
            A_tmp = np.dstack(A).sum(axis=2)
            B_tmp = np.dstack(B).sum(axis=2)
            C_tmp = np.dstack(C).sum(axis=2)

            if 'F' not in keep_param:
                # noinspection PyProtectedMember
                F = self.ssm_array[0]._m_update_f(A=A_tmp, B=B_tmp, priors=priors)
            else:
                F = self.ssm_array[0].F  # assume that all models already have the same F

            if 'Q' not in keep_param:
                # noinspection PyProtectedMember
                Q = self.ssm_array[0]._m_update_q(A=A_tmp, B=B_tmp, C=C_tmp, T=self.T, F=F, priors=priors)

        if update_mu0S0G:
            # noinspection PyUnboundLocalVariable
            C_tmp = np.dstack(C).sum(axis=2) if not update_FQ else C_tmp
            x_0_n_tmp = np.dstack(x_0_n).sum(axis=2) / self.K
            P_0_n_tmp = np.dstack(P_0_n).sum(axis=2) / self.K

            if 'mu0' not in keep_param:
                # noinspection PyProtectedMember
                mu0 = self.ssm_array[0]._m_update_mu0(x_0_n=x_0_n_tmp[:, 0])
            else:
                mu0 = self.ssm_array[0].mu0  # assume that all models already have the same mu0

            if 'S0' not in keep_param:
                # noinspection PyProtectedMember
                S0 = self.ssm_array[0]._m_update_S0(x_0_n=x_0_n_tmp[:, 0], P_0_n=P_0_n_tmp, mu0=mu0)

            if 'G' not in keep_param:
                # an additional sum of square needs to be calculated to update G
                D_tmp = np.dstack([self.ssm_array[k].y @ x_t_n_all[k][:, 1:].T for k in range(self.K)]).sum(axis=2)

                # noinspection PyProtectedMember
                G = self.ssm_array[0]._m_update_g(C=C_tmp, D=D_tmp)

        # Distribute to all models
        if update_FQ or update_mu0S0G:
            for k in range(self.K):
                if 'F' not in keep_param:
                    # noinspection PyUnboundLocalVariable
                    self.ssm_array[k].F = F

                if 'Q' not in keep_param:
                    # noinspection PyUnboundLocalVariable
                    self.ssm_array[k].Q = Q

                if 'mu0' not in keep_param:
                    # noinspection PyUnboundLocalVariable
                    self.ssm_array[k].mu0 = mu0

                if 'S0' not in keep_param:
                    # noinspection PyUnboundLocalVariable
                    self.ssm_array[k].S0 = S0

                if 'G' not in keep_param:
                    # noinspection PyUnboundLocalVariable
                    self.ssm_array[k].G = G if G is not None else self.ssm_array[k].G

                # make sure instance attributes reflect the updated parameters
                self.ssm_array[k].update_comp_param()



"""
Author: Mingjian He <mh1@stanford.edu>

ssm module contains a general state-space model class used in SOMATA
"""

from somata.exact_inference import kalman, djkalman, inverse
from somata.utils import estimate_r
import numpy as np
import numbers
from collections.abc import Iterable
from copy import deepcopy
from joblib import Parallel, delayed, cpu_count
from scipy.linalg import block_diag


class StateSpaceModel(object):
    """ StateSpaceModel is a general object class for state-space modeling in SOMATA """
    # Class attributes (should not be manually changed or be mutable)
    # only listed here as a reference list for all attributes under Ssm

    # Attributes are intentionally not implemented as properties with
    # decorators in order to allow for flexible input types. Type-hints
    # are omitted for the same reason. numpy.asanyarray() will handle
    # type conversions in the end.

    type = 'ssm'
    nstate = None
    ncomp = None
    nchannel = None
    ntime = None
    nmodel = None
    components = None
    comp_nstates = None
    F = None
    Q = None
    mu0 = None
    S0 = None
    G = None
    R = None
    y = None
    Fs = None
    stackable = ('F', 'Q', 'mu0', 'S0', 'G', 'R')
    default_G = None

    def __init__(self, components=None, F=None, Q=None, mu0=None, S0=None, G=None, R=None, y=None, Fs=None):
        """
        Constructor method for StateSpaceModel class
        :param components: a list of independent components in the model
        :param F: transition matrix
        :param Q: state noise covariance matrix
        :param mu0: initial state mean vector
        :param S0: initial state covariance matrix
        :param G: observation matrix (row major)
        :param R: observation noise covariance matrix
        :param y: observed data (row major, can be multivariate)
        :param Fs: sampling frequency in Hz
        """
        # Instance attributes
        self.F = self._process_constructor_input(F)
        self.Q = self._process_constructor_input(Q)
        self.mu0 = self._process_constructor_input(mu0)
        self.S0 = self._process_constructor_input(S0)
        self.y = self._must_be_row(self._process_constructor_input(y))
        self.Fs = np.float64(Fs) if Fs is not None else None

        # Initialize components
        if components is None:
            self.ncomp = 0
            self.components = None
            self.comp_nstates = [0]
        elif isinstance(components, str):
            self._auto_populate_components(components)
        else:
            self._initialize_from_components(components)

        # Initialize observation matrix and noise
        self._initialize_observation_matrix(G)
        self._initialize_observation_noise(R)

        # Check all dimensions
        self._check_dimensions()
        self._check_model_stack()

    # Dunder methods - magic methods and arithmetic-like operations
    def __repr__(self):
        """ Unambiguous and concise representation when calling StateSpaceModel() """
        return 'Ssm(' + str(self.ncomp) + ')<' + hex(id(self))[-4:] + '>'

    def __str__(self):
        """ Helpful information when calling print(StateSpaceModel()) """
        print_str = "<Ssm object at " + hex(id(self)) + ">\n " + \
                    "{0:8} = {1: <5} ".format("nstate", str(self.nstate)) + \
                    "{0:8} = {1}\n ".format("ncomp", str(self.ncomp)) + \
                    "{0:8} = {1: <5} ".format("nchannel", str(self.nchannel)) + \
                    "{0:8} = {1}\n ".format("ntime", str(self.ntime)) + \
                    "{0:8} = {1}\n ".format("nmodel", str(self.nmodel)) + \
                    "components = {}\n ".format(str(self.components)) + \
                    "{0:3}.shape = {1: <10} ".format("F", [str(x.shape) if x is not None else 'None' for x
                                                           in [self.F]][0]) + \
                    "{0:3}.shape = {1}\n ".format("Q", [str(x.shape) if x is not None else 'None' for x
                                                        in [self.Q]][0]) + \
                    "{0:3}.shape = {1: <10} ".format("mu0", [str(x.shape) if x is not None else 'None' for x
                                                             in [self.mu0]][0]) + \
                    "{0:3}.shape = {1}\n ".format("S0", [str(x.shape) if x is not None else 'None' for x
                                                         in [self.S0]][0]) + \
                    "{0:3}.shape = {1: <10} ".format("G", [str(x.shape) if x is not None else 'None' for x
                                                           in [self.G]][0]) + \
                    "{0:3}.shape = {1}\n ".format("R", [str(x.shape) if x is not None else 'None' for x
                                                        in [self.R]][0]) + \
                    "{0:3}.shape = {1: <10} ".format("y", [str(x.shape) if x is not None else 'None' for x
                                                           in [self.y]][0]) + \
                    "Fs = {0}\n ".format([str(x) + ' Hz' if x is not None else 'None' for x in [self.Fs]][0])
        return print_str

    def ssm(self):
        """ Create a StateSpaceModel object with the same attributes as the OscillatorModel object """
        return StateSpaceModel(components=self.components, F=self.F, Q=self.Q, mu0=self.mu0, S0=self.S0,
                               G=self.G, R=self.R, y=self.y, Fs=self.Fs)

    def __len__(self):
        return self.nmodel

    def __add__(self, other):
        return self._stack(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        return self._permute(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __matmul__(self, other):
        return self.__mul__(other)

    def __rmatmul__(self, other):
        return self.__matmul__(other)

    def __pow__(self, other: int):
        if other == 0:
            return StateSpaceModel()
        elif other > 0:
            # simply a magic method to call _expand_attr() with deepcopy()
            out_copy = deepcopy(self)
            out_copy._expand_attr()
            return out_copy
        else:
            raise SyntaxError('StateSpaceModel objects cannot be taken to negative power.')

    def __rpow__(self, other: int):
        return self.__pow__(other)

    # Check methods - sanity checks to ensure model validity
    def _check_dimensions(self, check_scope=(0, 1, 2)):
        """
        Check dimensions of instance attributes, ignoring None attributes
        :param check_scope: (int) specifying which ones to check.
            0 = state dimensions,
            1 = observed data dimensions,
            2 = components and observation matrix
        """
        if 0 in check_scope:
            # check the dimension of states
            check_state_tuple = (self.F, self.Q, self.mu0, self.S0)
            check_state_n = len(check_state_tuple)
            check_state_result = np.zeros(check_state_n, dtype=np.int_)
            for ii in range(check_state_n):
                if check_state_tuple[ii] is not None:
                    attr_shape = check_state_tuple[ii].shape
                    check_state_result[ii] = attr_shape[0]
                    if ii != 2:  # mu0 is a vector therefore no need to check
                        assert attr_shape[0] == attr_shape[1], 'Matrix attribute is not square!'

            nstates = np.unique(check_state_result[np.nonzero(check_state_result)])
            if len(nstates) == 0:
                self.nstate = 0
            else:
                assert len(nstates) == 1, 'Number of states is ambiguous and inconsistent across attributes.'
                self.nstate = int(nstates[0])

        if 1 in check_scope:
            # check the dimension of observation channels
            self.nchannel = 0
            self.ntime = 0
            if self.R is not None:
                assert self.R.shape[0] == self.R.shape[1], 'Matrix attribute is not square!'
                self.nchannel = self.R.shape[0]
            if self.y is not None:
                self.nchannel = self.y.shape[0]
                self.ntime = self.y.shape[1]
                if self.R is not None:
                    assert self.y.shape[0] == self.R.shape[0], \
                        'Observation dimensions are inconsistent across attributes.'

        if 2 in check_scope:
            # check the dimensions of components and observation matrix
            if self.components is not None:
                assert self.nstate == sum([x.default_G.shape[1] for x in self.components]), \
                    'Input components give a different state dimension from that of other attributes.'
                assert self.ncomp == len(self.components), 'Numbers of components are inconsistent.'
                self.comp_nstates = [x.default_G.shape[1] for x in self.components]
            else:
                self.comp_nstates = [0]

            if self.G is not None:
                if self.nchannel > 0:
                    assert self.G.shape[0] == self.nchannel, 'Observation matrix dimension mismatches nchannel.'
                if self.nstate > 0:
                    assert self.G.shape[1] == self.nstate, 'Observation matrix dimension mismatches nstate.'

    def _check_model_stack(self):
        """
        Check the number of alternative models stacked in instance attributes.
        third axis is reserved for stacking different models
        """
        # Different components and y should reside in different instances, therefore not stackable
        assert self.components is None or isinstance(self.components, list), \
            'Component structure should not be stacked in a single StateSpaceModel instance.'
        assert self.y is None or len(self.y.shape) < 3, \
            'Observed data y should not be stacked in a single StateSpaceModel instance.'
        assert self.Fs is None or isinstance(self.Fs, numbers.Number), \
            'Sampling frequency is not a valid Number type.'

        # The rest of instance attributes can be stacked with different models
        stack_numbers = self._get_stack_numbers()
        model_stack = np.unique(stack_numbers[stack_numbers != 1])
        if len(model_stack) == 0:
            self.nmodel = 1
        else:
            assert len(model_stack) == 1, 'More than one model stack numbers. Invoke __mul__ if trying to permute.'
            self.nmodel = int(model_stack[0])

    def _check_observed_data(self, other):
        """ Used to check if two objects have the same observed data """
        if self.y is not None and other.y is not None:
            assert self.y.shape == other.y.shape, 'Observed data have different dimensions.'
            # randomly select one time point to spot check
            rindex = np.random.randint(self.y.shape[0])
            cindex = np.random.randint(self.y.shape[1])
            assert self.y[rindex, cindex] == other.y[rindex, cindex], 'Observed data are not identical.'
        if self.Fs is not None and other.Fs is not None:
            assert self.Fs == other.Fs, 'Observed data have different sampling frequencies.'

    # Syntactic sugar methods - useful methods to make manipulations easier
    def _get_stack_numbers(self):
        """ Obtain the number of models stacked in each attribute """
        check_stack_tuple = tuple([getattr(self, x) for x in self.stackable])
        check_stack_n = len(check_stack_tuple)
        stack_numbers = np.ones(check_stack_n, dtype=np.int_)
        for ii in range(check_stack_n):
            if check_stack_tuple[ii] is not None:
                attr_shape = check_stack_tuple[ii].shape
                if len(attr_shape) > 2:
                    stack_numbers[ii] = attr_shape[2]
        return stack_numbers

    def _stack(self, other):
        """ Stack attributes in two objects """
        self._check_observed_data(other)
        new_obj = self._concat_attr(other)
        new_obj._check_model_stack()
        return new_obj

    def _permute(self, other):
        """ Permute attributes in two objects """
        self._check_observed_data(other)
        new_obj = self._concat_attr(other)
        new_obj._expand_attr()
        return new_obj

    def _concat_attr(self, other, attrs=None, ignore_duplicate=True):
        """
        Concatenate two objects via stacking new model attributes,
        duplicate checking is enabled; therefore repeated attributes
        will be ignored by turning add_flag to False, unless the
        ignore_duplicate flag is set to False
        """
        new_obj = deepcopy(self)
        if attrs is None:
            attrs = ('components', 'y', 'Fs') + self.stackable
        elif not isinstance(attrs, tuple):
            attrs = tuple(attrs)

        for attr_name in attrs:
            self_attr = getattr(self, attr_name)
            other_attr = getattr(other, attr_name)
            if self_attr is None and other_attr is None:
                pass
            elif other_attr is None:
                pass
            elif self_attr is None:
                setattr(new_obj, attr_name, deepcopy(other_attr))
            else:
                if attr_name == 'Fs':
                    assert self_attr == other_attr, 'Observed data have different sampling frequencies.'
                if attr_name == 'components':
                    assert len(self_attr) == len(other_attr), 'Components have different shapes.'
                elif len(other_attr.shape) > 2:
                    for j in range(other_attr.shape[2]):
                        model_attr = other_attr[:, :, j]
                        add_flag = True
                        if len(self_attr.shape) > 2:
                            for k in range(self_attr.shape[2]):
                                if (self_attr[:, :, k] == model_attr).all():  # type: ignore
                                    add_flag = False
                        else:
                            if (self_attr == model_attr).all():  # type: ignore
                                add_flag = False
                        if add_flag or ignore_duplicate is False:
                            setattr(new_obj, attr_name, np.dstack([getattr(new_obj, attr_name), model_attr]))
                else:
                    model_attr = other_attr
                    add_flag = True
                    if len(self_attr.shape) > 2:
                        for k in range(self_attr.shape[2]):
                            if (self_attr[:, :, k] == model_attr).all():  # type: ignore
                                add_flag = False
                    else:
                        if (self_attr == model_attr).all():  # type: ignore
                            add_flag = False
                    if add_flag or ignore_duplicate is False:
                        setattr(new_obj, attr_name, np.dstack([getattr(new_obj, attr_name), model_attr]))
        new_obj._check_dimensions()
        return new_obj

    def _expand_attr(self):
        """ Expand stacked attributes in an object """
        stack_numbers = self._get_stack_numbers()
        update_idx = [x for x in range(len(stack_numbers)) if (stack_numbers > 1)[x]]
        new_stack_number = int(stack_numbers[update_idx].prod())
        # expand with permutation of attributes to form stacked models
        last_attr_multiplier = 1
        for j in range(len(update_idx)):
            attr_name = self.stackable[update_idx[j]]
            current_attr = getattr(self, attr_name)
            rep_num = int(new_stack_number // last_attr_multiplier // stack_numbers[update_idx[j]])
            temp_attr = np.dstack([np.dstack([current_attr[:, :, n] for _ in range(rep_num)])
                                   for n in range(current_attr.shape[2])])
            setattr(self, attr_name, np.dstack([temp_attr for _ in range(last_attr_multiplier)]))
            last_attr_multiplier = int(last_attr_multiplier * stack_numbers[update_idx[j]])
        self._check_dimensions()
        self._check_model_stack()

    def _auto_populate_components(self, type_str):
        """ Auto-populate components with a given type of components """
        assert isinstance(type_str, str), 'components should be a string in this method.'
        self._check_dimensions((0,))  # fill in the nstate attribute

        if type_str == 'Osc':  # auto-populate with oscillator components
            self._auto_populate_osc_components()
        elif type_str == 'Arn':
            self._auto_populate_arn_components()
        elif type_str == 'Gen':
            self._auto_populate_gen_components()
        else:
            raise ValueError('Specified component type is not valid.')

        self._check_dimensions((2,))  # fill in the comp_nstates attribute

    def _auto_populate_osc_components(self):
        """ Initialize all components to be Matsuda oscillators during constructor """
        try:  # this has to stay here to avoid circular import before class definitions
            from .osc import OscillatorModel
        except (ImportError, ModuleNotFoundError):
            from somata import OscillatorModel

        assert (self.nstate % 2) == 0, 'Cannot default to Osc components with odd nstate.'
        self.ncomp = self.nstate // 2
        if self.ncomp > 0:
            self.components = [OscillatorModel() for _ in range(self.ncomp)]
        else:
            self.components = None

    def _auto_populate_arn_components(self):
        """ Initialize all components to be autoregressive models during constructor """
        try:  # this has to stay here to avoid circular import before class definitions
            from .arn import AutoRegModel
        except (ImportError, ModuleNotFoundError):
            from somata import AutoRegModel

        # F and Q parameters must be specified
        assert self.F is not None, 'F parameter is not available to form Arn components.'
        assert self.Q is not None, 'Q parameter is not available to form Arn components.'
        assert np.all(self.Q[~np.eye(self.Q.shape[0], dtype=bool)] == 0), \
            'Found non-zero non-diagonal elements when trying to auto-populate Arn components.'

        # Guess how many Arn components are contained in the model parameters
        Q_diag = np.diag(self.Q)
        comp_nstates = np.diff(np.append(np.nonzero(Q_diag), len(Q_diag)))
        assert sum(comp_nstates) == self.nstate, 'Arn components do not have the correct total number of states.'
        self.ncomp = len(comp_nstates)

        # Create the autoregressive model components
        self.components = []
        for n in range(len(comp_nstates)):
            start_idx = sum(comp_nstates[:n])
            end_idx = sum(comp_nstates[:n+1])
            coeff = self.F[start_idx, start_idx:end_idx]
            sigma2 = self.Q[start_idx, start_idx]
            mu0 = self.mu0[start_idx:end_idx] if self.mu0 is not None else None
            S0 = self.Q[start_idx, start_idx] if self.S0 is not None else None
            self.components.append(AutoRegModel(coeff=coeff, sigma2=sigma2, mu0=mu0, S0=S0))

    def _auto_populate_gen_components(self):
        """ Initialize with a single general state-space model during constructor """
        try:  # this has to stay here to avoid circular import before class definitions
            from .gen import GeneralSSModel
        except (ImportError, ModuleNotFoundError):
            from somata import GeneralSSModel

        if self.nstate > 0:
            self.ncomp = 1
            self.components = [GeneralSSModel()]
            self.components[0].default_G = np.ones((1, self.nstate), dtype=np.float64)
        else:
            self.ncomp = 0
            self.components = None

    def _initialize_from_components(self, components):
        """ Initialize the StateSpaceModel instance from given components during constructor """
        assert not isinstance(components, str), 'components should not be a string in this method.'

        # Form a list of single components without nesting
        components = list(components) if isinstance(components, Iterable) else [components]
        comp_list = []
        for n in range(len(components)):
            current_component = components[n]
            if current_component.ncomp <= 1:
                comp_list.append(current_component)
            else:
                assert isinstance(current_component.components, list), 'Invalid components attribute.'
                assert (np.asarray([x.ncomp for x in current_component.components]) <= 1).all(), \
                    'Components should not have nested multiple components in them.'
                comp_list += current_component.components

        self.ncomp = len(comp_list)
        self.components = comp_list
        self.comp_nstates = [x.default_G.shape[1] for x in self.components]

        # Use the Ssm.concat_() method to concatenate parameters from components
        concat_model = self.components[0]
        for n in range(1, self.ncomp):
            concat_model = StateSpaceModel.concat_(concat_model, self.components[n], skip_components=True)

        # Fill in model parameters if available from components
        self._setattr_when_not_none(concat_model, ('F', 'Q', 'mu0', 'S0', 'R', 'Fs'))

        # Fill in observed data if available from components
        self._check_observed_data(concat_model)
        self.y = self._return_not_none(self.y, concat_model.y)

    def _setattr_when_not_none(self, other, attrs):
        """ Update an instance attribute when the other model has non-empty values """
        attrs = tuple(attrs) if not isinstance(attrs, tuple) else attrs
        for attr_name in attrs:
            self_attr = getattr(self, attr_name)
            other_attr = getattr(other, attr_name)
            if other_attr is not None:
                if self_attr is None:
                    setattr(self, attr_name, other_attr)
                else:
                    assert np.all(self_attr == other_attr), \
                        'Incompatible attribute ' + attr_name + ' when initializing from components.'

    def _initialize_observation_matrix(self, G):
        """ Initialize the observation matrix instance attribute during constructor """
        if G is None:
            if self.ncomp > 0:
                nchannel = self.y.shape[0] if self.y is not None else 1  # default to one observation channel
                self.G = np.tile(self._must_be_row(self._process_constructor_input(
                    np.hstack([x.default_G for x in self.components]))), (nchannel, 1))
            else:
                self.G = None
        else:
            self.G = self._must_be_row(self._process_constructor_input(G))

    def _initialize_observation_noise(self, R):
        """ Initialize the observation noise covariance matrix during constructor """
        # Update R if not provided but can be estimated from y and Fs
        if R is None and self.y is not None and self.Fs is not None:
            R = estimate_r(y=self.y, Fs=self.Fs, freq_cutoff=self.Fs / 2 - 20)

        self.R = self._process_constructor_input(R)

    def stack_attr(self, attr_name, new_attr):
        """ Stack new_attr into the third axis of an instance attribute """
        if attr_name == 'components' or attr_name == 'y' or attr_name == 'Fs':
            raise AttributeError(attr_name + " shouldn't be stacked.")
        # using _concat_attr can automatically ignore duplicated attributes
        temp_obj = self._concat_attr(StateSpaceModel(**{attr_name: new_attr}), attrs=attr_name)
        setattr(self, attr_name, getattr(temp_obj, attr_name))
        self._check_dimensions()
        try:
            self._check_model_stack()
        except AssertionError:
            print("WARNING: nmodel cannot be updated automatically due to ambiguous model stack numbers.")

    def stack_to_array(self):
        """
        Unstack attributes into an array of objects,
        which is duck-typing-equivalent with a list
        """
        self._check_model_stack()
        stack_numbers = self._get_stack_numbers()
        update_idx = [x for x in range(len(stack_numbers)) if (stack_numbers > 1)[x]]
        ssm_array = np.empty(self.nmodel, dtype=StateSpaceModel)  # mutable array
        for ii in range(self.nmodel):
            temp_obj = deepcopy(self)
            for j in range(len(update_idx)):
                temp_attr = deepcopy(getattr(self, self.stackable[update_idx[j]]))
                setattr(temp_obj, self.stackable[update_idx[j]], temp_attr[:, :, ii])
            temp_obj._check_dimensions()
            temp_obj._check_model_stack()
            ssm_array[ii] = temp_obj
        return ssm_array

    def copy(self, drop_y=False):
        """ Make a copy of the StateSpaceModel Object """
        self_copy = deepcopy(self)
        if drop_y:
            self_copy.y = None
        return self_copy

    def append(self, other):
        """
        Append an object to an existing StateSpaceModel object,
        by calling concat_() and copying over attributes
        """
        self._setattr_for_append(temp_obj=self.concat_(other))

    def rappend(self, other):
        """
        Append an existing StateSpaceModel object to an object,
        positioned on the right side, i.e., ohter is appended from
        the left, by calling concat_() and copying over attributes
        """
        self._setattr_for_append(temp_obj=other.concat_(self))

    def _setattr_for_append(self, temp_obj):
        """
        Copy over attributes from temp_obj to self for appending.
        If self and other are of the same type, modify attributes
        in place, otherwise, raise an exception
        """
        if self.type == temp_obj.type:
            for attr in self.__dict__.keys():
                setattr(self, attr, getattr(temp_obj, attr))
        else:
            raise TypeError('Cannot append two SOMATA models of different classes. Use concat_() instead.')

    def concat_(self, other, skip_components=False):
        """
        Join two StateSpaceModel objects together by concatenating the
        components, and return a new object with the new total number
        of components. The mutable attribute components has new memory
        addresses for each component
        """
        # If both have observed data, they should be identical
        self._check_observed_data(other)
        y = self._return_not_none(self.y, other.y)
        Fs = self._return_not_none(self.Fs, other.Fs)
        if self.R is not None and other.R is not None:
            assert (self.R == other.R).all(), 'Cannot concatenate two objects with conflicting R.'  # type: ignore
            R = self.R
        else:
            R = self._return_not_none(self.R, other.R)

        # Fill components with parameters if concatenating with different classes of SOMATA basic models
        if self.type != other.type and not skip_components:
            components_prefill_self = self.fill_components()
            components_prefill_other = other.fill_components()

        # Set up the new components attribute using deepcopy(), without maintaining memory addresses
        new_comp = [None] * (self.ncomp + other.ncomp)
        if len(new_comp) == 0 or skip_components:
            new_comp = None
        else:
            new_comp[:self.ncomp] = [None] * self.ncomp if self.components is None else deepcopy(self.components)
            new_comp[self.ncomp:] = [None] * other.ncomp if other.components is None else deepcopy(other.components)

        # Revert any changes made to the components attribute of the original objects
        if self.type != other.type and not skip_components:
            self.unfill_components(components_prefill_self)
            other.unfill_components(components_prefill_other)

        # Configure the rest of attributes that are immutable
        # F
        if self.F is None and other.F is None:
            F = None
        elif self.F is None:
            F = block_diag(np.empty((self.nstate, self.nstate), dtype=object), other.F)
        elif other.F is None:
            F = block_diag(self.F, np.empty((other.nstate, other.nstate), dtype=object))
        else:
            F = block_diag(self.F, other.F)

        # Q
        if self.Q is None and other.Q is None:
            Q = None
        elif self.Q is None:
            Q = block_diag(np.empty((self.nstate, self.nstate), dtype=object), other.Q)
        elif other.Q is None:
            Q = block_diag(self.Q, np.empty((other.nstate, other.nstate), dtype=object))
        else:
            Q = block_diag(self.Q, other.Q)

        # mu0
        if self.mu0 is None and other.mu0 is None:
            mu0 = None
        elif self.mu0 is None:
            mu0 = np.vstack([np.zeros((self.nstate, 1), dtype=np.float64), other.mu0])
        elif other.mu0 is None:
            mu0 = np.vstack([self.mu0, np.zeros((other.nstate, 1), dtype=np.float64)])
        else:
            mu0 = np.vstack([self.mu0, other.mu0])

        # S0
        if self.S0 is None and other.S0 is None:
            S0 = None
        elif self.S0 is None:
            S0 = block_diag(self.Q, other.S0) if self.Q is not None else \
                block_diag(np.empty((self.nstate, self.nstate), dtype=object), other.S0)
        elif other.mu0 is None:
            S0 = block_diag(self.S0, other.Q) if other.Q is not None else \
                block_diag(self.S0, np.empty((other.nstate, other.nstate), dtype=object))
        else:
            S0 = block_diag(self.S0, other.S0)

        # G
        if self.G is None and other.G is None:
            G = None
        else:
            nchannel = y.shape[0] if y is not None else 1  # default to one observation channel

            if self.G is None:
                assert self.ncomp > 0, 'Cannot concatenate when self has no components.'
                G_self = np.tile(self._must_be_row(self._process_constructor_input(
                        np.hstack([x.default_G for x in self.components]))), (nchannel, 1))
            else:
                G_self = self.G

            if other.G is None:
                assert other.ncomp > 0, 'Cannot concatenate when other has no components.'
                G_other = np.tile(other._must_be_row(other._process_constructor_input(
                    np.hstack([x.default_G for x in other.components]))), (nchannel, 1))
            else:
                G_other = other.G

            G = np.hstack([G_self, G_other])

        return StateSpaceModel(components=new_comp, F=F, Q=Q, mu0=mu0, S0=S0, G=G, R=R, y=y, Fs=Fs)

    def remove_component(self, comp_idx):
        """ Remove a component from the Ssm object """
        if isinstance(comp_idx, list):  # recursive case
            comp_idx_list = list(np.sort(comp_idx)[::-1])
            for comp_idx in comp_idx_list:
                self.remove_component(comp_idx=comp_idx)
        else:  # base case
            start_idx = sum(self.comp_nstates[:comp_idx])
            end_idx = sum(self.comp_nstates[:comp_idx+1])
            slice_idx = range(start_idx, end_idx)
            self.nstate -= self.comp_nstates[comp_idx]
            self.ncomp -= 1
            _ = self.comp_nstates.pop(comp_idx)
            _ = self.components.pop(comp_idx)
            if self.F is not None:
                self.F = np.delete(self.F, slice_idx, axis=0)
                self.F = np.delete(self.F, slice_idx, axis=1)
            if self.Q is not None:
                self.Q = np.delete(self.Q, slice_idx, axis=0)
                self.Q = np.delete(self.Q, slice_idx, axis=1)
            if self.mu0 is not None:
                self.mu0 = np.delete(self.mu0, slice_idx, axis=0)
            if self.S0 is not None:
                self.S0 = np.delete(self.S0, slice_idx, axis=0)
                self.S0 = np.delete(self.S0, slice_idx, axis=1)
            if self.G is not None:
                self.G = np.delete(self.G, slice_idx, axis=1)

            # Set attributes to default values if nothing left
            self.comp_nstates = 0 if len(self.comp_nstates) == 0 else self.comp_nstates
            self.components = None if len(self.components) == 0 else self.components
            self.F = None if len(self.F) == 0 else self.F
            self.Q = None if len(self.Q) == 0 else self.Q
            self.mu0 = None if len(self.mu0) == 0 else self.mu0
            self.S0 = None if len(self.S0) == 0 else self.S0

    def fill_components(self, empty_comp=None, deep_copy=True):
        """
        Fill components with attribute content with deepcopy(),
        observed data y is copied using memory address to save space,
        so it remains mutable (be careful when you slice-alter y)
        """
        # Create an empty component instance
        empty_comp = empty_comp or StateSpaceModel()

        # Save the current components in case we need to revert to unfilled status
        components_prefill = deepcopy(self.components)

        for ii in range(self.ncomp):
            current_component: StateSpaceModel = self.components[ii]  # typehint to superclass to be flexible
            start_idx = sum(self.comp_nstates[:ii])
            end_idx = sum(self.comp_nstates[:ii+1])

            # Provide an empty component in the components attribute
            setattr(current_component, 'components', [deepcopy(empty_comp)])
            current_component.ncomp = 1
            current_component.components[0].default_G = current_component.default_G

            if getattr(self, 'F') is not None:
                if deep_copy:
                    F = deepcopy(getattr(self, 'F')[start_idx:end_idx, start_idx:end_idx])
                else:
                    F = getattr(self, 'F')[start_idx:end_idx, start_idx:end_idx]
                setattr(current_component, 'F', F)
            if getattr(self, 'Q') is not None:
                if deep_copy:
                    Q = deepcopy(getattr(self, 'Q')[start_idx:end_idx, start_idx:end_idx])
                else:
                    Q = getattr(self, 'Q')[start_idx:end_idx, start_idx:end_idx]
                setattr(current_component, 'Q', Q)
            if getattr(self, 'mu0') is not None:
                if deep_copy:
                    mu0 = deepcopy(getattr(self, 'mu0')[start_idx:end_idx])
                else:
                    mu0 = getattr(self, 'mu0')[start_idx:end_idx]
                setattr(current_component, 'mu0', mu0)
            if getattr(self, 'S0') is not None:
                if deep_copy:
                    S0 = deepcopy(getattr(self, 'S0')[start_idx:end_idx, start_idx:end_idx])
                else:
                    S0 = getattr(self, 'S0')[start_idx:end_idx, start_idx:end_idx]
                setattr(current_component, 'S0', S0)
            if getattr(self, 'G') is not None:
                if deep_copy:
                    G = deepcopy(getattr(self, 'G')[:, start_idx:end_idx])
                else:
                    G = getattr(self, 'G')[:, start_idx:end_idx]
                setattr(current_component, 'G', G)

            # R remains unchanged across components
            if deep_copy:
                setattr(current_component, 'R', deepcopy(getattr(self, 'R')))
            else:
                setattr(current_component, 'R', getattr(self, 'R'))

            setattr(current_component, 'y', getattr(self, 'y'))  # point to the same memory address
            setattr(current_component, 'Fs', getattr(self, 'Fs'))  # Fs should have immutable datatype

            # Use check methods in component subclasses to update dimension attributes
            current_component._check_dimensions()
            current_component._check_model_stack()

            # Update component parameters
            current_component.update_comp_param()

        return components_prefill

    def unfill_components(self, components_prefill):
        """ Unfill components to a prefill state """
        assert len(self.components) == len(components_prefill), 'Different numbers of components during unfill.'
        for ii in range(self.ncomp):
            attr_dict = components_prefill[ii].__dict__
            for attr in attr_dict.keys():
                setattr(self.components[ii], attr, getattr(components_prefill[ii], attr))

    def get_default_q(self, components=None, E=None):
        """
        Get the default structure of state noise covariance
        matrix Q in the Q_basis block diagonal form
        """
        components = self.components if components is None else components
        default_Q = block_diag(*[x.get_default_q(components=x, E=E) for x in components])
        return default_Q

    @staticmethod
    def _process_constructor_input(a):
        """
        Process inputs to the constructor of Ssm class.
        third axis is reserved for stacking different models instantiated with tuple inputs
        """
        if a is None:
            return a
        elif isinstance(a, tuple):
            return np.asanyarray(np.dstack(a), dtype=np.float64)
        else:
            # always break the link to input data memory address
            a = deepcopy(np.asanyarray(a, dtype=np.float64))

            if len(a.shape) == 0:  # always promote to (r,c) 2D arrays
                return a[None, None]
            elif len(a.shape) == 1:
                return a[:, None]
            else:
                return a

    @staticmethod
    def _must_be_row(x):
        """ Ensure input vector is a row vector """
        if x is None:
            return x
        elif x.shape[1] == 1 and len(x.shape) < 3:
            return x.T
        else:
            return x

    @staticmethod
    def _return_not_none(a, b):
        """ Return the first not-None value """
        if a is None:
            return b
        else:
            return a

    @staticmethod
    def setup_array(ssm_array, y=None, skip_check_observed=False):
        """ Prepare an array of Ssm objects for other signal processing methods """
        # Convert to an array if the input is a single Ssm instance
        if hasattr(ssm_array, 'nmodel'):  # suggests that this is a single somata class object
            ssm_array = ssm_array.stack_to_array()

        # Fill the models with the same observed data if provided
        if y is not None:
            for m in range(len(ssm_array)):
                ssm_array[m].y = y

        # Verify that observed data are identical across the array
        first_ssm: StateSpaceModel = ssm_array[0]
        if not skip_check_observed:
            for ii in range(1, len(ssm_array)):
                first_ssm._check_observed_data(ssm_array[ii])

        # Array dimensions
        K = len(ssm_array)  # number of models
        Ts = [x.ntime for x in ssm_array]  # number of time points
        if len(np.unique(Ts)) == 1:
            T = Ts[0]
        else:
            T = Ts  # allow variable lengths of data in different objects

        return ssm_array, K, T

    # Kalman filtering and smoothing methods (E step)
    def kalman_filt_smooth(self, y=None, R_weights=None, return_dict=False, EM=False, skip_interp=True, seterr=None):
        """ Wrapper method for classical kalman filtering and smoothing """
        if seterr is not None:  # apply np.seterr() for parallel processes
            old_settings = np.seterr(**seterr)

        y = self.y if y is None else y
        x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp = kalman(
            F=self.F, Q=self.Q, mu0=self.mu0, S0=self.S0, G=self.G, R=self.R, y=y, R_weights=R_weights,
            skip_interp=skip_interp)

        if seterr is not None:
            # noinspection PyUnboundLocalVariable
            np.seterr(**old_settings)

        if EM:  # return minimally necessary variables for EM algorithm
            return {'x_t_n': x_t_n, 'P_t_n': P_t_n, 'P_t_tmin1_n': P_t_tmin1_n, 'logL': logL}
        else:
            if return_dict:
                return {'x_t_n': x_t_n, 'P_t_n': P_t_n, 'P_t_tmin1_n': P_t_tmin1_n, 'logL': logL,
                        'x_t_t': x_t_t, 'P_t_t': P_t_t, 'K_t': K_t,
                        'x_t_tmin1': x_t_tmin1, 'P_t_tmin1': P_t_tmin1, 'fy_t_interp': fy_t_interp}
            else:
                return x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp

    def dejong_filt_smooth(self, y=None, R_weights=None, return_dict=False, EM=False, skip_interp=True, seterr=None):
        """ Wrapper method for De Jong version kalman filtering and smoothing """
        if seterr is not None:  # apply np.seterr() for parallel processes
            old_settings = np.seterr(**seterr)

        y = self.y if y is None else y
        x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp = djkalman(
            F=self.F, Q=self.Q, mu0=self.mu0, S0=self.S0, G=self.G, R=self.R, y=y, R_weights=R_weights,
            skip_interp=skip_interp)

        if seterr is not None:
            # noinspection PyUnboundLocalVariable
            np.seterr(**old_settings)

        if EM:  # return minimally necessary variables for EM algorithm
            return {'x_t_n': x_t_n, 'P_t_n': P_t_n, 'P_t_tmin1_n': P_t_tmin1_n, 'logL': logL}
        else:
            if return_dict:
                return {'x_t_n': x_t_n, 'P_t_n': P_t_n, 'P_t_tmin1_n': P_t_tmin1_n, 'logL': logL,
                        'x_t_t': x_t_t, 'P_t_t': P_t_t, 'K_t': K_t,
                        'x_t_tmin1': x_t_tmin1, 'P_t_tmin1': P_t_tmin1, 'fy_t_interp': fy_t_interp}
            else:
                return x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp

    @staticmethod
    def par_kalman(ssm_array, y=None, method='kalman', skip_check_observed=False,
                   R_weights=None, skip_interp=True, return_dict=False):
        """
        Parallel run kalman filtering and smoothing on
        an array of StateSpaceModel objects
        """
        ssm_array, K, _ = StateSpaceModel.setup_array(ssm_array, y=y, skip_check_observed=skip_check_observed)

        if method == 'kalman':
            kalman_func = getattr(StateSpaceModel, 'kalman_filt_smooth')
        elif method == 'dejong':
            kalman_func = getattr(StateSpaceModel, 'dejong_filt_smooth')
        else:
            raise ValueError('Specified method is invalid for parallel kalman calls.')

        if R_weights is None:
            R_weights = np.tile(np.array([None]), K)
        else:
            R_weights = tuple([R_weights[m, :] for m in range(K)])

        # Run kalman filtering and smoothing on parallel processes
        n_jobs = max(cpu_count()-1, 1)
        results = Parallel(n_jobs=n_jobs)(delayed(kalman_func)(model, y, weights, False, False,
                                                               skip_interp, np.geterr())
                                          for model, weights in zip(ssm_array, R_weights))

        # Unpack results into separate variables to return
        (x_t_n_all, P_t_n_all, P_t_tmin1_n_all,
         x_t_t_all, P_t_t_all, K_t_all, x_t_tmin1_all, P_t_tmin1_all,
         logL_all, fy_t_interp_all) = tuple([[None]*K for _ in range(10)])
        for m in range(K):
            (x_t_n_all[m], P_t_n_all[m], P_t_tmin1_n_all[m],
             logL_all[m], x_t_t_all[m], P_t_t_all[m], K_t_all[m],
             x_t_tmin1_all[m], P_t_tmin1_all[m], fy_t_interp_all[m]) = results[m]

        if return_dict:
            return {'x_t_n_all': x_t_n_all, 'P_t_n_all': P_t_n_all, 'P_t_tmin1_n_all': P_t_tmin1_n_all,
                    'logL_all': logL_all, 'x_t_t_all': x_t_t_all, 'P_t_t_all': P_t_t_all, 'K_t_all': K_t_all,
                    'x_t_tmin1_all': x_t_tmin1_all, 'P_t_tmin1_all': P_t_tmin1_all, 'fy_t_interp_all': fy_t_interp_all}
        else:
            return x_t_n_all, P_t_n_all, P_t_tmin1_n_all, logL_all, \
                x_t_t_all, P_t_t_all, K_t_all, x_t_tmin1_all, P_t_tmin1_all, fy_t_interp_all

    # Parameter estimation methods (M step)
    # noinspection PyUnusedLocal
    def m_estimate(self, y=None, x_t_n=None, P_t_n=None, P_t_tmin1_n=None, h_t=None, logL=None,
                   priors=None, A=None, B=None, C=None, T=None, force_ABC=False,
                   update_param=('F', 'Q', 'mu0', 'S0', 'G', 'R'), keep_param=(),
                   return_dict=None):
        """
        Maximum likelihood or Maximum a posteriori estimation to update
        parameters. Ssm class doesn't have explicit state-equation parameter
        update rules, therefore calling the inherited _m_update_<param>
        methods in component subclasses

        Reference:
            Shumway, R. H., & Stoffer, D. S. (1982). An approach to time series
            smoothing and forecasting using the EM algorithm. Journal of time series
            analysis, 3(4), 253-264.

            Ghahramani, Z., & Hinton, G. E. (1996). Parameter estimation for linear
            dynamical systems.

            Ghahramani, Z., & Hinton, G. E. (2000). Variational learning for
            switching state-space models. Neural computation, 12(4), 831-864.

        Inputs:
        :param self: Ssm class instance
        :param y: observed data
        :param x_t_n: smoothed estimates (posterior) of state mean
        :param P_t_n: smoothed estimates (posterior) of state conditional covariance
        :param P_t_tmin1_n: smoothed estimates (posterior) of state lag1 conditional cross-covariance
        :param h_t: responsibility vector (i.e. 1/R_weights in E step)
        :param logL: unused variable, pure syntactic sugar to allow pythonic **d inputs
        :param priors: a list of dictionaries specifying priors for each component, if None -> MLE
        :param A: sum of square terms
        :param B: sum of square terms
        :param C: sum of square terms
        :param T: number of time points in sum of square terms
        :param force_ABC: flag to compute ABC regardless of state-equation parameters
        :param update_param: a tuple of strings for parameters to update
        :param keep_param: a tuple of strings for parameters to keep and not update
        :param return_dict: None -> no return, True -> return dict, False -> return tuple of variables
        """
        # Obtain boolean flags of scopes of updates
        update_FQ, update_mu0S0G = self._m_estimate_scope(update_param, keep_param)

        # Initialize parameters
        y = self.y if y is None else y
        if priors is None:
            priors = [None] * self.ncomp
        elif isinstance(priors, dict):
            priors = [priors]  # so that it can be indexed at position 0

        # Attempt to skip sums of squares computation
        if update_FQ or force_ABC:
            A = self._m_compute_ss('A', x_t_n, P_t_n=P_t_n) if A is None else A
            B = self._m_compute_ss('B', x_t_n, P_t_tmin1_n=P_t_tmin1_n) if B is None else B
            C = self._m_compute_ss('C', x_t_n, P_t_n=P_t_n) if C is None else C
            T = x_t_n.shape[1] - 1 if T is None else T

        # Initialize the responsibility vector h_t
        if h_t is None:
            h_t_length = T if T is not None else y.shape[1]
            h_t = np.ones(h_t_length, dtype=np.float64)  # default responsibility is 1 for all timepoints

        # Update parameters for each independent component -- F, Q, mu0, S0, G (component specific priors)
        if update_FQ or update_mu0S0G:
            for ii in range(self.ncomp):  # iterate through components
                current_component: StateSpaceModel = self.components[ii]  # typehint to superclass to be flexible
                start_idx = sum(self.comp_nstates[:ii])
                end_idx = sum(self.comp_nstates[:ii+1])
                if update_FQ:
                    A_tmp = A[start_idx:end_idx, start_idx:end_idx]
                    B_tmp = B[start_idx:end_idx, start_idx:end_idx]
                    C_tmp = C[start_idx:end_idx, start_idx:end_idx]
                else:
                    A_tmp, B_tmp, C_tmp = (None, None, None)

                # Call the _m_update_<param> methods specific to the component subclass
                if 'F' in update_param and 'F' not in keep_param:
                    self.F[start_idx:end_idx, start_idx:end_idx] = \
                        current_component._m_update_f(A=A_tmp, B=B_tmp, priors=priors[ii])

                if 'Q' in update_param and 'Q' not in keep_param:
                    self.Q[start_idx:end_idx, start_idx:end_idx] = \
                        current_component._m_update_q(A=A_tmp, B=B_tmp, C=C_tmp, T=T,
                                                      F=self.F[start_idx:end_idx, start_idx:end_idx],
                                                      priors=priors[ii])

                if 'mu0' in update_param and 'mu0' not in keep_param:
                    self.mu0[start_idx:end_idx, 0] = \
                        current_component._m_update_mu0(x_0_n=x_t_n[start_idx:end_idx, 0])[:, 0]

                if 'S0' in update_param and 'S0' not in keep_param:
                    self.S0[start_idx:end_idx, start_idx:end_idx] = \
                        current_component._m_update_S0(x_0_n=x_t_n[start_idx:end_idx, 0],
                                                       P_0_n=P_t_n[start_idx:end_idx, start_idx:end_idx, 0],
                                                       mu0=self.mu0[start_idx:end_idx, 0][:, None])

                if 'G' in update_param and 'G' not in keep_param:
                    comp_G = current_component._m_update_g(y=y, x_t_n=x_t_n[start_idx:end_idx, :],
                                                           P_t_n=P_t_n[start_idx:end_idx, start_idx:end_idx, :],
                                                           h_t=h_t)
                    self.G[:, start_idx:end_idx] = comp_G if comp_G is not None else self.G[:, start_idx:end_idx]

        # Update observation noise covariance -- R
        if 'R' in update_param and 'R' not in keep_param:
            R_ss = self._m_update_r(y=y, x_t_n=x_t_n, P_t_n=P_t_n, h_t=h_t, G=self.G, priors=priors[0])
        else:
            R_ss = self._m_update_r(y=y, x_t_n=x_t_n, P_t_n=P_t_n, h_t=h_t, G=self.G, keep_R=True)

        if return_dict is None:
            pass
        elif return_dict:
            return {'F': self.F, 'Q': self.Q, 'mu0': self.mu0, 'S0': self.S0, 'G': self.G, 'R': self.R,
                    'R_ss': R_ss, 'A': A, 'B': B, 'C': C}
        else:
            return self.F, self.Q, self.mu0, self.S0, self.G, self.R, R_ss, A, B, C

    def update_comp_param(self):
        """ Update component specific parameters, override in subclasses """
        pass

    def e_step(self, y=None, logL_list=None, **kwargs):
        """ Exposed E step method for run_em() """
        e_results: dict = self.dejong_filt_smooth(y=y, EM=True, **kwargs)
        e_logL = e_results['logL'].sum()
        stop_var = e_logL - logL_list[-1]  # logL should monotonically increase
        logL_list.append(e_logL)
        return e_results, stop_var

    def m_step(self, y=None, **kwargs):
        """ Exposed M step method for run_em() """
        self.m_estimate(y=y, **kwargs)
        return

    def initialize_priors(self):
        """ Initialize priors for each component in the object """
        assert self.components is not None, 'Cannot initialize priors when components is None.'
        components_prefill = self.fill_components()
        priors = [x.initialize_priors() for x in self.components]
        self.unfill_components(components_prefill)
        return priors

    def _initialize_priors_recursive_list(self, prior_value):
        """ Expand a prior value to the length of components for the recursive case """
        if prior_value is None:
            prior_value = [None] * self.ncomp
        elif isinstance(prior_value, str) or not isinstance(prior_value, Iterable) or len(prior_value) == 1:
            prior_value = [prior_value] * self.ncomp
        else:
            assert len(prior_value) == self.ncomp, 'Different lengths of specified prior value and components.'
        return prior_value

    def _m_update_r(self, y=None, x_t_n=None, P_t_n=None, h_t=None, G=None, R_ss=None, T=None, priors=None,
                    keep_R=False):
        """ Update observation noise covariance """
        if R_ss is None:
            assert len(h_t) == y.shape[1], 'Different lengths of h_t and y. Cannot proceed with _m_update_r().'
            R_ss = (h_t * (y - G @ x_t_n[:, 1:])) @ (y - G @ x_t_n[:, 1:]).T + \
                G @ (h_t * P_t_n[:, :, 1:]).sum(axis=2) @ G.T

        if not keep_R:  # proceed to update self.R
            T = h_t.sum() if T is None else T

            if self._m_update_if_mle('R_sigma2', priors):
                # MLE
                self.R = R_ss / T
            else:
                # MAP with inverse gamma prior
                R_init = priors['R_sigma2']
                R_hp = priors['R_hyperparameter'] if 'R_hyperparameter' in priors else 0.1
                alpha = T * R_hp / 2  # scales with data length T according to the hyperparameter
                beta = R_init * (alpha + 1)  # setting the mode of inverse gamma prior to be R_init
                self.R = (beta + R_ss/2) / (alpha + T/2 + 1)  # new R is the mode of inverse gamma posterior

        return R_ss

    @staticmethod
    def _m_update_if_mle(params, priors):
        """ Check whether doing MLE during _m_update_<param> methods """
        if priors is None:
            return True

        if isinstance(params, str) or not isinstance(params, Iterable):
            params = [params]

        MLE_flag = np.zeros(len(params), dtype=bool)
        for param, idx in zip(params, range(len(params))):
            if param not in priors:
                MLE_flag[idx] = True
            elif isinstance(priors[param], str):
                if priors[param].upper() == 'MLE':
                    MLE_flag[idx] = True

        return MLE_flag.all()

    # noinspection PyUnusedLocal
    @staticmethod
    def _m_update_f(A=None, B=None, priors=None):
        pass

    # noinspection PyUnusedLocal
    @staticmethod
    def _m_update_q(A=None, B=None, C=None, T=None, F=None, priors=None):
        pass

    # noinspection PyUnusedLocal
    @staticmethod
    def _m_update_q_src(Q_ss=None, T=None, Q_basis=None, nsource=None, nstate=None, priors=None):
        pass

    @staticmethod
    def _m_update_mu0(x_0_n=None):
        """ Update initial state mean -- mu0 """
        mu0 = x_0_n[:, None].copy()
        return mu0

    @staticmethod
    def _m_update_mu0_src(x_0_n=None, nstate=None):
        """
        Update initial state mean mu0 in dynamic
        source localization
        """
        assert x_0_n.shape[0] % nstate == 0, 'Input x_t_n is not a multiple of nstate.'
        mu0 = x_0_n[:, None].clone()
        return mu0

    @staticmethod
    def _m_update_S0(x_0_n=None, P_0_n=None, mu0=None):
        """ Update initial state covariance -- S0 """
        S0 = P_0_n + x_0_n[:, None] @ x_0_n[:, None].T \
            - x_0_n[:, None] @ mu0.T - mu0 @ x_0_n[:, None].T + mu0 @ mu0.T
        return S0

    @staticmethod
    def _m_update_g(y=None, x_t_n=None, P_t_n=None, h_t=None, C=None, D=None):
        """ Update observation matrix -- G """
        if C is None:
            assert len(h_t) == y.shape[1], 'Different lengths of h_t and y. Cannot proceed with _m_update_g().'
            C = (h_t * P_t_n[:, :, 1:]).sum(axis=2) + (h_t * x_t_n[:, 1:]) @ x_t_n[:, 1:].T

        if D is None:
            assert len(h_t) == y.shape[1], 'Different lengths of h_t and y. Cannot proceed with _m_update_g().'
            D = (h_t * y) @ x_t_n[:, 1:].T

        approach = 'svd' if C.shape[0] >= 5 else 'gaussian'
        G = D @ inverse(C, approach=approach)
        return G

    @staticmethod
    def _m_estimate_scope(update_param=('F', 'Q', 'mu0', 'S0', 'G', 'R'), keep_param=()):
        """ Sort out the scopes of updates in m_estimate() """
        if 'F' not in update_param and 'Q' not in update_param:
            update_FQ = False
        elif 'F' in keep_param and 'Q' in keep_param:
            update_FQ = False
        else:
            update_FQ = True

        if 'mu0' not in update_param and 'S0' not in update_param and 'G' not in update_param:
            update_mu0S0G = False
        elif 'mu0' in keep_param and 'S0' in keep_param and 'G' in keep_param:
            update_mu0S0G = False
        else:
            update_mu0S0G = True

        return update_FQ, update_mu0S0G

    @staticmethod
    def _m_compute_ss(ss_term, x_t_n, P_t_n=None, P_t_tmin1_n=None):
        """
        Compute the sum of squares terms for m_estimate().
        Definitions of A,B,C follow the notations in equations (9,10,11)
        of S&S 1982, and these terms also correspond to maximizing <H>_Q
        in G&H 2000.

        Reference:
            Shumway, R. H., & Stoffer, D. S. (1982). An approach to time series
            smoothing and forecasting using the EM algorithm. Journal of time series
            analysis, 3(4), 253-264.

            Ghahramani, Z., & Hinton, G. E. (2000). Variational learning for
            switching state-space models. Neural computation, 12(4), 831-864.
        """
        if ss_term == 'A':
            return P_t_n[:, :, :-1].sum(axis=2) + x_t_n[:, :-1] @ x_t_n[:, :-1].T
        elif ss_term == 'B':
            return P_t_tmin1_n[:, :, 1:].sum(axis=2) + x_t_n[:, 1:] @ x_t_n[:, :-1].T
        elif ss_term == 'C':
            return P_t_n[:, :, 1:].sum(axis=2) + x_t_n[:, 1:] @ x_t_n[:, 1:].T
        else:
            raise ValueError('Invalid ss_term is specified.')



"""
Author: Mingjian He <mh1@stanford.edu>

gen module contains general Gaussian state-space model methods used in SOMATA
"""

from somata.basic_models import StateSpaceModel as Ssm
from somata.exact_inference import inverse
import numpy as np
from scipy.linalg import block_diag


class GeneralSSModel(Ssm):
    """
    GeneralSSModel is a subclass of StateSpaceModel class dedicated
    to a general single component structure of linear Gaussian
    state-space models
    """
    type = 'gen'
    default_G = None

    def __init__(self, components='Gen', F=None, Q=None, mu0=None, S0=None, G=None, R=None, y=None, Fs=None):
        """
        Constructor method for GeneralSSModel class
        :param components: add a single component or None
        :param F: transition matrix
        :param Q: state noise covariance matrix
        :param mu0: initial state mean vector
        :param S0: initial state covariance matrix
        :param G: observation matrix (row major)
        :param R: observation noise covariance matrix
        :param y: observed data (row major, can be multivariate)
        :param Fs: sampling frequency in Hz
        """
        # Call parent class constructor
        super().__init__(components=components, F=F, Q=Q, mu0=mu0, S0=S0, G=G, R=R, y=y, Fs=Fs)
        self.default_G = np.ones((1, self.nstate), dtype=np.float64)

    # Dunder methods - magic methods
    def __repr__(self):
        """ Unambiguous and concise representation when calling GeneralSSModel() """
        return super().__repr__().replace('Ssm', 'Gen')

    def __str__(self):
        """ Helpful information when calling print(GeneralSSModel()) """
        print_str = super().__str__().replace('<Ssm object at', '<Gen object at')
        return print_str

    # Syntactic sugar methods - useful methods to make manipulations easier
    def concat_(self, other, skip_components=False):
        """
        Join two GeneralSSModel objects together by concatenating the
        components.
        """
        assert self.type == 'gen', 'self in concat_() needs to be of GeneralSSModel class.'

        if other.type == self.type:  # concatenation within GeneralSSModel class
            # Call parent class method to obtain general concatenated attributes
            temp_obj = super().concat_(other, skip_components=skip_components)
            new_obj = GeneralSSModel(components=temp_obj.components, F=temp_obj.F, Q=temp_obj.Q,
                                     mu0=temp_obj.mu0, S0=temp_obj.S0,
                                     G=temp_obj.G, R=temp_obj.R, y=temp_obj.y, Fs=temp_obj.Fs)

        else:  # concatenation with other classes of SOMATA basic models
            new_obj = super().concat_(other, skip_components=skip_components)

        return new_obj

    def get_default_q(self, components=None, E=None):
        """
        Get the default structure of state noise covariance
        matrix Q in the Q_basis block diagonal form
        """
        components = self.components if components is None else components
        assert len(components) == 1, 'More than one component for GeneralSSModel class is not permitted.'
        first_component = components[0]
        assert isinstance(first_component, GeneralSSModel), 'The only component type is not GeneralSSModel.'

        order = first_component.default_G.shape[1]
        E = np.eye(1, dtype=np.float64) if E is None else E
        default_Q = block_diag(*[E] * order)
        return default_Q

    def fill_components(self, empty_comp=None, deep_copy=True):
        """ Fill the components attribute with GeneralSSModel parameters """
        empty_comp = empty_comp or GeneralSSModel()
        return super().fill_components(empty_comp=empty_comp, deep_copy=deep_copy)

    # Parameter estimation methods (M step)
    def initialize_priors(self, R_sigma2=None, R_hyperparameter=None):
        """ Initialize priors for general state-space component """
        assert self.ncomp <= 1, 'GeneralSSModel object should not have multiple components.'

        # [Inverse gamma prior] on observation noise variance R <--- TODO: update to Wishart
        if R_sigma2 is None:
            if self.R.shape[0] > 1:
                raise NotImplementedError('Only uni-variate observation data is supported with prior for now.')
            else:
                R_sigma2 = self.R[0, 0]
                R_hyperparameter = 0.1 if R_hyperparameter is None else R_hyperparameter

        return {'R_sigma2': R_sigma2, 'R_hyperparameter': R_hyperparameter}

    @staticmethod
    def _m_update_f(A=None, B=None, priors=None):
        """ Update transition matrix -- F """
        approach = 'svd' if A.shape[0] >= 5 else 'gaussian'
        F = B @ inverse(A, approach=approach)
        return F

    @staticmethod
    def _m_update_q(A=None, B=None, C=None, T=None, F=None, priors=None):
        """ Update state noise covariance matrix -- Q """
        Q = (C - B @ F.T - F @ B.T + F @ A @ F.T) / T
        return Q

    @staticmethod
    def _m_update_q_src(Q_ss=None, T=None, Q_basis=None, nsource=None, nstate=None, priors=None):
        """
        Update state noise covariance matrix Q in dynamic
        source localization, in the Q_basis block diagonal
        form
        """
        import torch
        from ..source_loc import SourceLocModel as Src
        assert Q_basis.shape == Q_ss.shape, 'GeneralSSModel has non-diagonal entries of component Q, ' \
                                            'therefore Q_basis must span the expanded space instead of nsource.'
        assert Q_ss.shape[0] == nsource * nstate, 'Q_ss does not match the dimension of GeneralSSModel with '\
                                                  + str(nstate) + ' states across ' + str(nsource) + ' sources.'

        # GeneralSSModel has a full matrix of sum of squares
        if isinstance(Q_basis, list):  # using non-orthonormal kernel
            Theta = Src.update_theta(Q_ss=Q_ss, T=T, Q_basis=Q_basis[1],
                                     nsource=nsource, npart=1, priors=priors)
            Q_new = Q_basis[0] @ torch.diag(Theta) @ Q_basis[0].T
        else:  # using orthonormal basis
            Theta = Src.update_theta(Q_ss=Q_ss, T=T, Q_basis=Q_basis,
                                     nsource=nsource, npart=1, priors=priors)
            Q_new = Q_basis @ torch.diag(Theta) @ Q_basis.T
        return Q_new



"""
Author: Mingjian He <mh1@stanford.edu>

Testing functions for the dOsc algorithm in somata/oscillator_search/decomp_osc.py
"""

from somata.oscillator_search import DecomposedOscillatorModel as DecOsc
from somata.oscillator_search.helper_functions import (
    random, np, simulate_matsuda, sim_to_osc_object)
from test_load_data import _load_data  # type: ignore
import pickle


def test_decomposed_osc(plot_on=False):
    random.seed(1)
    fs = 100
    osc1 = {'a': 0.996, 'q': 0.4, 'f': 0.1}
    osc2 = {'a': 0.95, 'q': 0.2, 'f': 10}
    y, param_list, ob_noise = simulate_matsuda([osc1, osc2], R=1.2, Fs=fs, T=10)
    sim_osc0, sim_x0 = sim_to_osc_object(y, param_list)

    do_test = DecOsc(y, fs)
    do_test.iterate(plot_fit=plot_on)

    # Verify diagnostic statistical tests
    acf = do_test.diagnose_residual_acf()
    assert 1.9 < acf['Durbin-Watson d'].item() < 2.1, "Failed Durbin-Watson test."
    assert acf['Ljung-Box'].item() > 0.05, "Failed Ljung-Box test."
    assert acf['Lagrange Multiplier'].item() > 0.05, "Failed Lagrange Multiplier test."
    assert acf['Brock-Dechert-Scheinkman'].item() > 0.05, "Failed Brock-Dechert-Scheinkman test."

    norm = do_test.diagnose_residual_norm()
    assert norm['Shapiro-Wilk'].item() > 0.05, "Failed Shapiro-Wilk test."
    assert norm['D\'Agostino and Pearson'].item() > 0.05, "Failed D'Agostino and Pearson test."
    assert norm['Anderson-Darling H'].item() is False, "Failed Anderson-Darling test."
    assert norm['Cramer-von Mises'].item() > 0.05, "Failed Cramer-von Mises test."

    if plot_on:
        # Plot frequency domain (from parameters and from estimated x_t_n)
        for version in ['theoretical', 'actual']:
            _ = do_test.get_knee_osc().visualize_freq(version, y=y, sim_osc=sim_osc0, sim_x=sim_x0)

        # Plot time domain estimated x_t
        _ = do_test.get_knee_osc().visualize_time(y=y, sim_x=sim_x0)

        # Plot likelihood and selected model (may not be the highest likelihood)
        _ = do_test.plot_log_likelihoods()

        # Plot multitaper spectrogram and mean spectrum
        _ = do_test.plot_mtm()

        # Plot raw time trace
        _ = do_test.plot_trace()

        # Plot fitted spectra (equivalent to calling .visualize_freq(version) manually)
        _ = do_test.plot_fit_spectra()

        # Plot the residual spectrum shifted by white noise processes
        _ = do_test.plot_residual()

        # Plot residual linear line fit
        _ = do_test.plot_residual_fit()

        # Plot autocorrelation function of the residuals
        _ = do_test.plot_acf()

        # Plot partial autocorrelation function of the residuals
        _ = do_test.plot_pacf()

    with open(_load_data('test_dec_osc_obj.pkl', return_path=True), 'rb') as inp:
        do_true = pickle.load(inp)

    assert do_true.knee_index == do_test.knee_index, "Does not choose the same oscillator."
    assert len(do_true.fitted_osc) == len(do_test.fitted_osc), "Does not have the same number of maximum oscillators."
    assert np.allclose(do_true.ll, do_test.ll), 'Log-likelihoods are not the same.'

    osc_true = do_test.get_knee_osc()
    osc_orig = do_test.get_knee_osc()

    assert np.allclose(osc_true.freq, osc_orig.freq), "Frequencies in DecOsc are not the same."
    assert np.allclose(osc_true.a, osc_orig.a), "Radii in DecOsc are not the same."
    assert np.allclose(osc_true.sigma2, osc_orig.sigma2), "sigma^2 in DecOsc are not the same."
    assert np.allclose(osc_true.R, osc_orig.R), "R (obs noise) in DecOsc are not the same."


if __name__ == "__main__":
    test_decomposed_osc()
    print('Decomposed oscillator test finished without exception.')



"""
Author: Ran Liu <rliu20@stanford.edu>

Unit tests for functions in somata/pac/pac_model.py
"""

import numpy as np
from somata.pac.pac_model import (
    fit_pac_regression,
    kmod,
    autocovariances,
    block_toeplitz,
    ar_parameters,
    mvar_ssm,
    optimize_arp,
)


def generate_pac_data():
    """ Generate test data for pac regression (not stochastic)"""
    n = int(1e3)
    Fs = 100
    phase = np.arange(0, n / Fs * 2 * np.pi, 2 * np.pi / Fs)
    amplitude = np.sin(phase)
    amplitude[amplitude < 0] = 0

    return phase, amplitude


def generate_ar3_data(rng=np.random.default_rng(0), n=int(1e3)):
    """ Generate test data from a univariate AR(3) process """
    p = 3  # autoregressive order of 3
    x0 = np.array([1., 2., 3.])
    F = np.array([[1, -0.5, 0.3], [1., 0., 0.], [0., 1., 0.]])  # hard-coded for AR(3)

    x = np.zeros((p, n), dtype=np.float64)  # augmented hidden states
    current = x0

    for i in range(n):
        current = F @ current
        current[0] += rng.standard_normal()  # Q = 1.0
        x[:, i] = current

    y = x[0, :] + rng.standard_normal(n)  # R = 1.0; G = [1, 0, 0]

    return x, y


def generate_mvar_data(q, rng=np.random.default_rng(0), n=int(1e3)):
    """
    Generate test data from a multivariate AR(3) process

    :param q: number of observed variables (dimensionality of the mvAR process)
    :param rng: random number generator
    :param n: length of the mvAR time series
    """
    data = [generate_ar3_data(rng) for i in range(q)]  # univariate AR(3) data for each variable
    x = np.row_stack([x[0, :] for x, _ in data])
    y = np.row_stack([y for _, y in data])  # make a mvAR(3) from three univariate AR(3) data

    return x, y


def test_pac_regression():
    phase, amplitude = generate_pac_data()

    # test whether the constrained regression is fitted correctly
    fit = fit_pac_regression(phase, amplitude, verbose=False)
    posterior_mean = fit.mean(0)  # posterior mean as an approximate to posterior mode

    fitted_kmod = kmod(posterior_mean[0], posterior_mean[1], posterior_mean[2])

    assert fitted_kmod >= 0, "Estimated Kmod should be non-negative."
    assert fitted_kmod < 1, "Estimated Kmod should be less than 1."


def test_autocovariances():
    rng = np.random.default_rng(0)
    q = 2  # mvAR with 2 variables
    p = 2  # up to lag 2

    _, y = generate_mvar_data(q=q, rng=rng)
    acov = autocovariances(y, p, bias=True)

    # Assert that we have p + 1 covariance matrices
    assert len(acov) == p + 1, 'Incorrect number of autocovariance matrices.'

    n = y.shape[1]
    for i in range(len(acov)):
        # Assert that each covariance matrix is of the right shape
        assert np.array_equal(acov[i].shape, (q, q))

        for j in range(q):
            for k in range(q):
                # Assert that each entry is the correct value
                assert np.isclose(
                    acov[i][j, k], np.cov(y[j, :(n - i)], y[k, i:n], bias=True)[0, 1]
                ), 'Incorrect autocovariance value.'


def test_block_toeplitz():
    rng = np.random.default_rng(0)
    q = 5  # mvAR with 5 variables
    p = 4  # up to lag 4

    _, y = generate_mvar_data(q=q, rng=rng)
    acov = autocovariances(y, p, bias=True)

    C = block_toeplitz(acov)

    # Assert that the block-toeplitz matrix is of the right shape
    assert np.array_equal(C.shape, ((p + 1) * q, (p + 1) * q)), 'Incorrect shape of C.'

    # Assert that the block-toeplitz matrix is symmetric
    assert np.allclose(C, C.T), 'C is not symmetric.'  # use np.allclose() due to numerical precision errors in np.cov()

    # Assert that the block-toeplitz matrix is positive definite
    assert np.all(np.linalg.eigvals(C) > 0), 'C is not positive definite.'

    for i in range(p + 1):
        for j in range(i + 1):
            # Assert diagonal and lower triangular matrices
            assert np.array_equal(
                C[(i * q):((i + 1) * q), (j * q):((j + 1) * q)], acov[i-j]
            ), 'Incorrect block-toeplitz matrix in diagonal or lower triangular blocks.'
        for j in range(i + 1, p + 1):
            # Assert upper triangular matrices
            assert np.array_equal(
                C[(i * q):((i + 1) * q), (j * q):((j + 1) * q)], acov[j-i].T,
            ), 'Incorrect block-toeplitz matrix in upper triangular blocks.'


def test_ar_parameters():
    rng = np.random.default_rng(0)
    q = 3  # mvAR with 3 variables
    p = 3  # up to lag 3, same as the generative AR(3) process

    _, y = generate_mvar_data(q=q, rng=rng)
    acov = autocovariances(y, p, bias=True)

    # get "approximate ground truth" based on the first univariate AR(3) variable
    A_hat, Q_hat = ar_parameters(autocovariances(y[0, :][None, :], p, bias=True), R=1.)
    A_hat = np.squeeze(np.concatenate(A_hat))

    A, Q = ar_parameters(acov, R=1.)

    assert len(A) == p, 'Incorrect number of AR parameter matrices.'
    for i in range(p):
        assert np.array_equal(A[i].shape, (q, q)), 'Incorrect shape of AR parameter matrix.'

    a = np.array([[A_k[i, i] for A_k in A] for i in range(q)])  # diagonal elements of A matrices

    assert len(a) == q, 'Incorrect number of mvAR variables.'
    for j in range(q):
        assert np.argmax(a[j]) == np.argmax(A_hat), 'Incorrect leading AR coefficient.'
        for i in range(p):
            assert np.sign(a[j][i]) == np.sign(A_hat[i]), 'Incorrect sign of AR coefficients.'

    assert np.allclose(Q, Q.T), 'Q is not symmetric.'  # use np.allclose() due to numerical precision errors
    assert ((np.abs(np.diag(Q) - Q_hat) / Q_hat) < 0.25).all(), 'Incorrect diagonal elements of Q.'


def test_mvar_ssm():
    rng = np.random.default_rng(0)
    q = 3  # mvAR with 3 variables

    x, y = generate_mvar_data(q=q, rng=rng)
    obs_error = np.mean((x - y) ** 2)

    for p in range(1, 6):
        acov = autocovariances(y, p, bias=True)
        R = 1.
        A, Q = ar_parameters(acov, R=R)

        model = mvar_ssm(y, A, Q, R)
        z = model.kalman_filt_smooth(return_dict=True)

        filtered_x = z["x_t_t"][:q, 1:]
        smoothed_x = z["x_t_n"][:q, 1:]

        filtered_error = np.mean((x - filtered_x) ** 2)
        smoothed_error = np.mean((x - smoothed_x) ** 2)

        assert filtered_error < obs_error, "Filtered error should be less than observation error."
        assert smoothed_error < filtered_error, "Smoothed error should be less than filtered error."


def test_optimize_arp():
    rng = np.random.default_rng(0)
    q = 3  # mvAR with 3 variables
    _, y = generate_mvar_data(q=q, rng=rng)
    bic = []

    for p in range(1, 6):
        A, Q, R = optimize_arp(y, p=p)
        assert R > 0 and not np.isclose(R, 0), 'Observation noise variance R should be positive.'
        assert (np.diag(Q) > 0).all(), 'Diagonal elements of state noise covariance Q should be positive.'
        assert not np.isclose(Q, np.zeros_like(Q)).any(), 'State noise covariance Q should not be zero.'

        model = mvar_ssm(y, A, Q, R)
        z = model.kalman_filt_smooth(return_dict=True)
        bic.append(-2 * z['logL'].sum() + (q * q * (p + 1) + 1) * np.log(y.shape[1]))

    assert np.argmin(bic) == 0, 'Optimal mvAR order should be 1.'


if __name__ == "__main__":
    test_pac_regression()
    test_autocovariances()
    test_block_toeplitz()
    test_ar_parameters()
    test_mvar_ssm()
    test_optimize_arp()



"""
Authors: Mingjian He <mh1@stanford.edu>
         Proloy Das <pd640@mgh.harvard.edu>

source_loc_utils module implements supporting functions for dynamic source localization
"""

import torch
import mne
import numpy as np
from somata.exact_inference import djkalman_conv_torch


def patch_decompose(fwd, src_patch, dist_metric='geodesic', normalize=False):
    """
    Patch decomposition for constructing the forward model of a
    source space with reduced dimensionality.

    Reference:
        Limpiti, T., Van Veen, B. D., & Wakai, R. T. (2006). Cortical patch
        basis model for spatially extended neural activity. IEEE Transactions
        on Biomedical Engineering, 53(9), 1740-1754.

        Babadi, B., Obregon-Henao, G., Lamus, C., Hämäläinen, M. S.,
        Brown, E. N., & Purdon, P. L. (2014). A subspace pursuit-based
        iterative greedy hierarchical solution to the neuromagnetic inverse
        problem. NeuroImage, 87, 427-443.

    Inputs:
    :param fwd: MNE Forward instance with source information
    :param src_patch: MNE SourceSpaces instance with reduced dimensionality
                      - Note: the recommended way to construct src_patch is
                      to use the same SourceSpaces construction pipeline in
                      creating fwd and use a smaller number of sources. If
                      the surfaces in fwd are simply decimated, there is no
                      guarantee that the smaller source space is a subset
                      of the one with higher dimension
    :param dist_metric: the distance metric to define Voronoi cell.
                        Options: 'geodesic' or 'Euclidean'
    :param normalize: whether to normalize the lead field matrix to have unit norms
    """
    # Make sure both SourceSpaces instances have the same coordinate frame
    src = fwd['src']
    assert src_patch[0]['coord_frame'] == fwd['coord_frame'] == src[0]['coord_frame'], \
        'Different coordinate frames between SourceSpaces.'

    # Ensure fixed normal direction is used for lead field gain matrix
    if fwd['sol']['data'].shape == fwd['_orig_sol'].shape:
        fwd = mne.convert_forward_solution(fwd, force_fixed=True, surf_ori=True)

    # Compute the pairwise-distance between all sources in use
    if dist_metric == 'geodesic' and src[0]['dist'] is None:
        mne.add_source_space_distances(src)

    # Process one hemisphere at a time
    lfm_patch_hemi = [np.array([None])] * len(src_patch)
    sval_hemi = [np.array([None])] * len(src_patch)
    nmra_hemi = [np.array([None])] * len(src_patch)
    for hemi in range(len(src_patch)):
        # Since the sources included in a patch are selected based on the Voronoi Cell
        # on the cortical surface of the source space with higher dimension, each patch
        # model source must also be a source in the source space with higher dimension
        assert src_patch[hemi]['np'] == src[hemi]['np'], 'Mismatch in the number of vertices.'
        assert src_patch[hemi]['ntri'] == src[hemi]['ntri'], 'Mismatch in the number of faces.'
        assert np.all(src[hemi]['inuse'][src_patch[hemi]['vertno']] == 1), 'Patch is not a subset of sources.'

        # Find the Voronoi cell of patch sources
        if dist_metric == 'geodesic':
            # using geodesic distances over cortical surface
            dist = src[hemi]['dist'][np.tile(src_patch[hemi]['vertno'], reps=(len(src[hemi]['vertno']), 1)).T,
                                     np.tile(src[hemi]['vertno'], reps=(len(src_patch[hemi]['vertno']), 1))].toarray()
            min_indices = np.argmin(dist, axis=0)  # Voronoi cell source indices
        else:
            # using Euclidean distances over 3D space
            min_indices = np.zeros(len(src[hemi]['vertno']), dtype=np.int32)
            for jj in range(len(min_indices)):
                vec_diff = np.linalg.norm(src[hemi]['rr'][src[hemi]['vertno'][jj], :]
                                          - src_patch[hemi]['rr'][src_patch[hemi]['vertno'], :], axis=1)
                min_indices[jj] = np.argmin(vec_diff)

        # Initialize the patch basis forward model lead field matrix
        lfm_hemi = fwd['sol']['data'][:, :src[0]['nuse']] if hemi == 0 else fwd['sol']['data'][:, src[0]['nuse']:]
        lfm_patch_hemi[hemi] = np.zeros((lfm_hemi.shape[0], src_patch[hemi]['nuse']))
        nmra_hemi[hemi] = np.zeros(lfm_patch_hemi[hemi].shape[1])  # normalized mean representation accuracy

        # Loop through each patch
        for ii in range(lfm_patch_hemi[hemi].shape[1]):
            # SVD of the patch forward model gain matrix
            U, s, *_ = np.linalg.svd(lfm_hemi[:, min_indices == ii])
            if normalize:
                lfm_patch_hemi[hemi][:, ii] = U[:, 0]  # first left singular vector
            else:
                lfm_patch_hemi[hemi][:, ii] = U[:, 0] * s[0]  # multiplied by first singular value

            sval_hemi[hemi][ii] = s[0]  # store the first singular value
            nmra_hemi[hemi][ii] = s[0] / s.sum()  # first singular value over the sum of all singular values

    # Concatenate the lead field matrices from two hemispheres
    lfm_patch = np.hstack(lfm_patch_hemi)
    sval = np.hstack(sval_hemi)
    nmra = np.hstack(nmra_hemi)

    # Note that the patch basis model does not support free orientation,
    # therefore we modify the '_orig_sol' and '_orig_source_ori' attributes
    # in the output fwd_patch instance accordingly to support saving with
    # mne.write_forward_solution() and reading with mne.read_forward_solution()
    fwd_patch = fwd.copy()
    fwd_patch['source_ori'] = mne.io.constants.FIFF.FIFFV_MNE_FIXED_ORI
    fwd_patch['nsource'] = lfm_patch.shape[1]
    fwd_patch['sol']['ncol'] = lfm_patch.shape[1]
    fwd_patch['sol']['data'] = lfm_patch
    fwd_patch['_orig_sol'] = lfm_patch
    fwd_patch['src'] = src_patch.copy()
    fwd_patch['source_rr'] = np.vstack(
        [src_patch[hemi]['rr'][src_patch[hemi]['vertno'], :] for hemi in range(len(src_patch))])
    fwd_patch['_orig_source_ori'] = mne.io.constants.FIFF.FIFFV_MNE_FIXED_ORI
    fwd_patch['source_nn'] = np.vstack(
        [src_patch[hemi]['nn'][src_patch[hemi]['vertno'], :] for hemi in range(len(src_patch))])

    return fwd_patch, sval, nmra


def simulate_oscillation(center_frequency_Hz, damping_factor, state_noise_variance, initial_state_mean,
                         initial_state_variance, sampling_frequency, duration, oscillation_type='oscillator'):
    """ Simulate oscillation activity time series """
    from somata.basic_models import OscillatorModel as Osc

    time_step = 1 / sampling_frequency
    time_axis = np.arange(0, duration + time_step, time_step)
    n_time_steps = time_axis.shape[0]

    if oscillation_type == 'oscillator':
        transition_matrix = Osc.get_rot_mat(damping_factor, Osc.hz_to_rad(center_frequency_Hz, sampling_frequency))
        simulated = np.zeros((2, n_time_steps))
        simulated[:, 0] += np.matmul(transition_matrix,
                                     np.random.multivariate_normal(initial_state_mean,
                                                                   initial_state_variance * np.eye(2, 2))
                                     ) + np.random.multivariate_normal([0, 0], state_noise_variance * np.eye(2, 2))
        for ii in range(n_time_steps - 1):
            simulated[:, ii + 1] = np.matmul(transition_matrix, simulated[:, ii]) + \
                                   np.random.multivariate_normal([0, 0], state_noise_variance * np.eye(2, 2))
        simulated = simulated[1, :]  # keep only the real component

    elif oscillation_type == 'sinusoid':
        # amplitude = 2 * np.pi * damping_factor * np.sqrt(state_noise_variance)  # this might not be correct
        # simulated = amplitude * np.sin(2 * np.pi * center_frequency_Hz * time_axis) + \
        #             np.random.normal(0, np.sqrt(state_noise_variance), n_time_steps)
        simulated = np.sin(2 * np.pi * center_frequency_Hz * time_axis)

    else:
        raise ValueError('Unrecognized oscillation_type value')

    return simulated


def resolution_matrix_metrics(K, src):
    """
    Compute the resolution matrix metrics:
        - Spatial dispersion (SD)
        - Dipole localization error (DLE)
        - Resolution index (RI)
    """
    from somata.source_loc import SourceLocModel as Src

    nsources = K.shape[0]

    # Compute the pairwise-distance between all triangulation vertices
    if src[0]['dist'] is None:
        mne.add_source_space_distances(src)

    # Create a distance matrix between sources
    D = np.zeros((nsources, nsources), dtype=np.float64)
    # noinspection PyProtectedMember
    _, source_to_vert = Src._vertex_source_mapping(src)

    for hemi in range(len(src)):
        for vidx in source_to_vert[hemi]:
            vert = source_to_vert[hemi][vidx]  # vertex indexing

            for vidx2 in source_to_vert[hemi]:
                vert2 = source_to_vert[hemi][vidx2]  # vertex indexing

                D[vidx, vidx2] = src[hemi]['dist'][vert, vert2] * 100  # get into cm

    # Calculate SD, DLE, RI
    SD = np.zeros(nsources, dtype=np.float64)
    DLE = np.zeros(nsources, dtype=np.float64)
    RI = np.zeros(nsources, dtype=np.float64)
    max_dist = np.max(D)

    for hemi in range(len(src)):
        for vidx in source_to_vert[hemi]:

            vidx2_list = np.asarray(list(source_to_vert[hemi].keys()))

            numerator = ((D[vidx, vidx2_list] * K[vidx2_list, vidx]) ** 2).sum()
            denominator = (K[vidx2_list, vidx] ** 2).sum()
            SD[vidx] = np.sqrt(numerator / denominator)

            max_r_idx = np.argmax(abs(K[vidx2_list, vidx]))
            DLE[vidx] = D[vidx2_list[max_r_idx], vidx]

            max_c_idx = np.argmax(abs(K[vidx, vidx2_list]))
            RI[vidx] = ((max_dist - D[vidx, vidx2_list[max_c_idx]]) * abs(K[vidx, vidx])
                        ) / (max_dist * abs(K[vidx, vidx2_list[max_c_idx]]))

    return SD, DLE, RI


def smart_djkalman_conv_torch(F, Q, mu0, S0, G, R, y, rank, proj, conv_steps=100):
    """
    Wrapper of djkalman_conv_torch() to pre-whiten observed data

    `smart_djkalman_conv_torch()` handles average referencing by combining it with
    whitening. This also handles the rank deficiency that comes with average
    referencing. Following is a brief explanation of why and how:

    djkalman_conv_torch() uses the following state space model:
        x(t) = F*x(t-1)+eta(t)   eta ~ N(0,Q) (state or transition equation)
        y(t) = G*x(t)+eps(t)     eps ~ N(0,R) (observation or measurement equation)

    In EEG, such y(t) is not observable because electrical potentials can only be
    measured upto a baseline (i.e, with recording referencing). Consider average
    referencing is captured by a matrix H, the actual observed data is then:
        z(t) = H*y(t) = H*(G*x(t) + eps(t))

    smart_djkalman_conv_torch() adds a whitening matrix that includes the referencing
    matrix to the observation equation:
        W*y(t) = W*(G*x(t)+eps(t))      where W = M*pinv(H), and M*R*M.T = I
               = W*G*x(t)+lambda(t)     lambda(t) = W*eps(t) ~ N(0,C)

    Because H is rank deficient, C is also rank-deficient. However, C is only
    involved in djkalman_conv_torch() when computing the Kalman gain:
        K = F*P*G.T*W.T(W*G*P*G.T*W.T + C)^{-1}     where C = W*R*W.T

    We can apply the Woodbury matrix identity (matrix inversion lemma) twice in
    terms that involve the Kalman gain:
        K(W*y - F*x)    when updating the mean
        K*W*G           when updating the covariance

    and replace C with an identity matrix. It is ok to use an identity matrix to
    replace a singular identity matrix (some diagonal values being zero) because
    the whitener W will preserve the rank-deficiency in these terms.

    Input/Output:
    ------------
    rank: int
         estimated rank of *projected* data
    proj: Ny x Ny matrix
         projection matrix computed from `mne.Info` object (use accompanied `make_projector()`
         to compute the matrix)

    see the rest in function header of djkalman_conv_torch()
    """
    dtype = y.dtype

    W, logdet_by_n2 = get_whitener(R=R.cpu().numpy(), rank=rank, proj=proj)  # W = M * pinv(H)
    W = torch.as_tensor(W, dtype=dtype).cuda()
    logdet_by_n2 = torch.as_tensor(logdet_by_n2, dtype=dtype).cuda()

    G = torch.matmul(W, G)
    y = torch.matmul(W, y)
    R = torch.eye(R.shape[0], dtype=dtype).cuda()

    x_t_n, P_t_n, P_t_tmin1_n, logL, break_conv, K_t, x_t_tmin1, P_t_tmin1 = djkalman_conv_torch(
        F, Q, mu0, S0, G, R, y, conv_steps, dtype)

    # adjust log-likelihood by adding back the -1/2 logdet term
    qlog2pi_by_n2 = -y.shape[0] * torch.log(torch.as_tensor(2 * torch.pi, dtype=dtype).cuda()) / 2
    logL += y.shape[1] * (qlog2pi_by_n2 + logdet_by_n2)

    return x_t_n, P_t_n, P_t_tmin1_n, logL, break_conv, K_t, x_t_tmin1, P_t_tmin1


def get_whitener(R, rank, proj):
    """ Compute a whitening matrix based on the observation noise covariance """
    # Apply the projector on observation noise covariance matrix
    R = proj @ R @ proj.T

    # Eigen-decomposition for whitening
    eig, eigvec = mne.utils.linalg.eigh(R, overwrite_a=True)
    eig[:-rank] = 0.0
    eig_pos_idx = eig > 0
    eig[~eig_pos_idx] = 0.0

    # Some sanity checks on eigenvalues
    assert eig[-1] > 0, 'The largest eigenvalue is non-positive.'
    assert np.all(eig[eig_pos_idx] / eig[-1] > 1e-15), \
        'Some eigenvalues are very small. Whitener matrix may be inaccurate.'

    # Create the whitener matrix
    eig[eig_pos_idx] = 1.0 / np.sqrt(eig[eig_pos_idx])
    logdet_by_n2 = np.log(eig[eig_pos_idx]).sum()
    W = eigvec * eig @ eigvec.T

    return W, logdet_by_n2


def make_projector(info, projs=None, ch_names=None):
    """
    Create the projection operator

    Author: Proloy Das <pd640@mgh.harvard.edu>

    Input:
    -----
    info: mne.Info
    projs : list | None (default)
        List of projection vectors.
    ch_names: list of str | None (default)
        List of channels to include in the projection matrix.
    """
    from mne.io.pick import pick_info
    from mne.io.proj import make_projector

    projs = info['projs'] if projs is None else projs
    ch_names = info['ch_names'] if ch_names is None else ch_names
    if info['ch_names'] != ch_names:
        info = pick_info(info, [info['ch_names'].index(c) for c in ch_names])
    assert info['ch_names'] == ch_names

    proj, ncomp, _ = make_projector(projs, ch_names)
    return proj, ncomp
